# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os

import tensorflow as tf
from absl import app, logging

import neurst.utils.flags_core as flags_core
from neurst.utils.compat import wrapper_var_name
from neurst.utils.configurable import ModelConfigs
from neurst.utils.flags_core import Flag
from neurst.utils.misc import flatten_string_list

FLAG_LIST = [
    flags_core.Flag("checkpoints", dtype=Flag.TYPE.STRING, default=None, multiple=True,
                    help="A list or comma-separated string of checkpoints to be averaged. "
                         "The averaged checkpoint will be saved to `output_path`."),
    flags_core.Flag("output_path", dtype=flags_core.Flag.TYPE.STRING,
                    help="The path to the averaged checkpoint."),
]


def checkpoint_exists(path):
    return (tf.io.gfile.exists(path) or tf.io.gfile.exists(path + ".meta")
            or tf.io.gfile.exists(path + ".index"))


def checkpoint_list_checking(path_list):
    if path_list:
        new_path_list = []
        for path in path_list:
            if checkpoint_exists(path):
                new_path_list.append(path)
        return new_path_list
    return []


def average_checkpoints(checkpoints, output_path):
    assert checkpoints
    # Get the checkpoints list from flags and run some basic checks.
    checkpoints = flatten_string_list(checkpoints)
    checkpoints = [c for c in checkpoints if c]
    if not checkpoints:
        raise ValueError("No checkpoints provided for averaging.")
    model_config_yml_path = None
    for c in checkpoints:
        if model_config_yml_path:
            break
        if tf.io.gfile.exists(os.path.join(c, ModelConfigs.MODEL_CONFIG_YAML_FILENAME)):
            model_config_yml_path = os.path.join(c, ModelConfigs.MODEL_CONFIG_YAML_FILENAME)
    all_checkpoint_paths = []
    for c in checkpoints:
        if tf.io.gfile.isdir(c):
            checkpoint_states = tf.train.get_checkpoint_state(c)
            all_checkpoint_paths.extend(checkpoint_list_checking(checkpoint_states.all_model_checkpoint_paths))
        else:
            all_checkpoint_paths.append(c)
    var_values = {}
    var_cnts = {}
    var_name_shape_list = tf.train.list_variables(all_checkpoint_paths[0])
    for ckpt in all_checkpoint_paths:
        logging.info("loading from {}".format(ckpt))
        for var_name, _ in var_name_shape_list:
            if var_name.startswith("_") or var_name.startswith("save_counter"):
                logging.info("ignore {}...".format(var_name))
                continue
            var = tf.train.load_variable(ckpt, var_name)
            fine_name = wrapper_var_name(var_name)
            if fine_name in var_values:
                var_cnts[fine_name] += 1.
                var_values[fine_name] = var * 1. / var_cnts[fine_name] + var_values[fine_name] * (
                    var_cnts[fine_name] - 1.) / var_cnts[fine_name]
            else:
                var_cnts[fine_name] = 1.
                var_values[fine_name] = var
    tf_vars = dict()
    logging.info("Averaged variables: ")
    for var_name in var_values.keys():
        fine_name = wrapper_var_name(var_name)
        assert var_cnts[fine_name] == len(all_checkpoint_paths)
        logging.info(fine_name)
        tf_vars[fine_name] = tf.Variable(
            initial_value=var_values[fine_name],
            trainable=True,
            name=fine_name,
            dtype=str(var_values[fine_name].dtype))
    ckpt_saver = tf.train.Checkpoint(**tf_vars)
    ckpt_saver.save(os.path.join(output_path, "ckpt"))
    tf.io.gfile.copy(model_config_yml_path,
                     os.path.join(output_path, ModelConfigs.MODEL_CONFIG_YAML_FILENAME),
                     overwrite=True)


def _main(_):
    arg_parser = flags_core.define_flags(FLAG_LIST, with_config_file=False)
    args, remaining_argv = flags_core.intelligent_parse_flags(FLAG_LIST, arg_parser)
    flags_core.verbose_flags(FLAG_LIST, args, remaining_argv)
    average_checkpoints(
        checkpoints=flatten_string_list(args["checkpoints"]),
        output_path=args["output_path"])


if __name__ == "__main__":
    logging.set_verbosity(logging.INFO)
    app.run(_main, argv=["pseudo.py"])



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import tensorflow as tf
from absl import app, logging

import neurst.utils.flags_core as flags_core
from neurst.data.datasets import Dataset, build_dataset
from neurst.data.datasets.audio.audio_dataset import RawAudioDataset

FLAG_LIST = [
    flags_core.Flag("output_transcript_file", dtype=flags_core.Flag.TYPE.STRING,
                    required=True, help="The path to save transcriptions."),
    flags_core.Flag("output_translation_file", dtype=flags_core.Flag.TYPE.STRING,
                    default=None, help="The path to save transcriptions."),
    flags_core.ModuleFlag(Dataset.REGISTRY_NAME, help="The raw dataset."),
]


def main(dataset, output_transcript_file, output_translation_file=None):
    assert isinstance(dataset, RawAudioDataset)
    transcripts = dataset.transcripts
    translations = dataset.translations
    assert transcripts, "Fail to extract transcripts."
    with tf.io.gfile.GFile(output_transcript_file, "w") as fw:
        fw.write("\n".join(transcripts) + "\n")
    if translations and output_translation_file:
        with tf.io.gfile.GFile(output_translation_file, "w") as fw:
            fw.write("\n".join(translations) + "\n")


def _main(_):
    # define and parse program flags
    arg_parser = flags_core.define_flags(FLAG_LIST, with_config_file=True)
    args, remaining_argv = flags_core.intelligent_parse_flags(FLAG_LIST, arg_parser)
    flags_core.verbose_flags(FLAG_LIST, args, remaining_argv)
    dataset = build_dataset(args)
    if dataset is None:
        raise ValueError("dataset must be provided.")
    main(dataset=dataset,
         output_transcript_file=args["output_transcript_file"],
         output_translation_file=args["output_translation_file"])


if __name__ == "__main__":
    logging.set_verbosity(logging.INFO)
    app.run(_main, argv=["pseudo.py"])



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import random

import numpy
import tensorflow as tf
from absl import app, logging

import neurst.utils.flags_core as flags_core
from neurst.data.datasets import Dataset, build_dataset
from neurst.tasks import Task, build_task
from neurst.utils.compat import ModeKeys

FLAG_LIST = [
    flags_core.Flag("processor_id", dtype=flags_core.Flag.TYPE.INTEGER, default=None,
                    help="The processor id."),
    flags_core.Flag("num_processors", dtype=flags_core.Flag.TYPE.INTEGER, default=None,
                    help="The number of processors. Must be divisible by `num_output_shards`."),
    flags_core.Flag("num_output_shards", dtype=flags_core.Flag.TYPE.INTEGER, default=None,
                    help="The total number of output shards."),
    flags_core.Flag("output_range_begin", dtype=flags_core.Flag.TYPE.INTEGER, default=None,
                    help="The begin ID of output shard (startswith 0, inclusive)."),
    flags_core.Flag("output_range_end", dtype=flags_core.Flag.TYPE.INTEGER, default=None,
                    help="The end ID of output shard (startswith 0, exclusive)."),
    flags_core.Flag("output_template", dtype=flags_core.Flag.TYPE.STRING, default="train.tfrecords-%5.5d-of-%5.5d",
                    help="The template name of output tfrecords, like train.tfrecords-%5.5d-of-%5.5d."),
    flags_core.Flag("mode", dtype=flags_core.Flag.TYPE.STRING, default=ModeKeys.TRAIN,
                    choices=ModeKeys._fields,  # pylint: disable=protected-access
                    help="The mode to acquire data preprocess method, "
                         "that is, the result TF Record dataset will be used for."),
    flags_core.Flag("progressbar", dtype=flags_core.Flag.TYPE.BOOLEAN, default=None,
                    help="Whether to dispaly the progressbar"),
    flags_core.ModuleFlag(Task.REGISTRY_NAME, help="The binding task for data pre-processing."),
    flags_core.ModuleFlag(Dataset.REGISTRY_NAME, help="The raw dataset."),
]


def _format_tf_feature(feature, dtype):
    if dtype is str:
        feature = tf.nest.map_structure(lambda _x: _x.encode("utf-8"), feature)
    value = numpy.array(feature).flatten()
    if dtype is int:
        return tf.train.Feature(int64_list=tf.train.Int64List(value=value))
    elif dtype is float:
        return tf.train.Feature(float_list=tf.train.FloatList(value=value))
    return tf.train.Feature(bytes_list=tf.train.BytesList(value=value))


def main(processor_id, num_processors,
         num_output_shards, output_range_begin, output_range_end,
         output_template, dataset: Dataset, progressbar=False, task=None):
    assert 0 <= output_range_begin < output_range_end <= num_output_shards
    assert 0 <= processor_id < num_processors
    logging.info(f"Shards: {output_range_begin} to {output_range_end}")
    if not tf.io.gfile.exists(os.path.dirname(output_template)):
        tf.io.gfile.makedirs(os.path.dirname(output_template))

    file_paths = [output_template % (s, num_output_shards) for s
                  in range(output_range_begin, output_range_end)]
    tmp_file_paths = [f + ".incomplete" for f in file_paths]
    recordio_writers = [tf.io.TFRecordWriter(_x) for _x in tmp_file_paths]

    map_func = None
    if task is not None:
        map_func = task.get_data_preprocess_fn(ModeKeys.TRAIN, dataset.status)

    feature_type_dict = None
    i = 0
    if progressbar:
        from tqdm import tqdm
        iterator = tqdm(dataset.build_iterator(map_func=map_func, shard_id=processor_id,
                                               total_shards=num_processors)(),
                        total=dataset.num_samples // num_processors)
    else:
        iterator = dataset.build_iterator(
            map_func=map_func, shard_id=processor_id, total_shards=num_processors)()
    for example in iterator:  # lazily pre-process
        if feature_type_dict is None:
            feature_type_dict = dict()
            for name, data in example.items():
                data_type = type(numpy.array(data).flatten().tolist()[0])
                assert data_type in [int, float, str, bytes], "Not supported {}".format(data_type)
                feature_type_dict[name] = data_type
        feature_dict = {}
        for name, data in example.items():
            feature_dict[name] = _format_tf_feature(data, feature_type_dict[name])
        recordio_writers[random.randint(0, len(recordio_writers) - 1)].write(
            tf.train.Example(features=tf.train.Features(feature=feature_dict)).SerializeToString())
        i += 1
    logging.info(f"Total processed {i} samples.")
    for recordio_writer in recordio_writers:
        recordio_writer.close()
    for tmp_f, f in zip(tmp_file_paths, file_paths):
        tf.io.gfile.rename(tmp_f, f, overwrite=True)
    logging.info("===================== Examine feature types =====================")
    for x in tf.data.TFRecordDataset(file_paths).take(1):
        example = tf.train.Example()
        example.ParseFromString(x.numpy())
        logging.info("{")
        for name in example.features.feature:
            if len(example.features.feature[name].bytes_list.value) > 0:
                logging.info(f"    \"{name}\": bytes (str)")
            elif len(example.features.feature[name].int64_list.value) > 0:
                logging.info(f"    \"{name}\": int64")
            elif len(example.features.feature[name].float_list.value) > 0:
                logging.info(f"    \"{name}\": float32")
        logging.info("}")


def _main(_):
    # define and parse program flags
    arg_parser = flags_core.define_flags(FLAG_LIST, with_config_file=True)
    args, remaining_argv = flags_core.intelligent_parse_flags(FLAG_LIST, arg_parser)
    flags_core.verbose_flags(FLAG_LIST, args, remaining_argv)
    task = build_task(args)
    dataset = build_dataset(args)
    if dataset is None:
        raise ValueError("dataset must be provided.")
    main(processor_id=args["processor_id"],
         num_processors=args["num_processors"],
         num_output_shards=args["num_output_shards"],
         output_range_begin=args["output_range_begin"],
         output_range_end=args["output_range_end"],
         output_template=args["output_template"],
         progressbar=args["progressbar"],
         dataset=dataset, task=task)


if __name__ == "__main__":
    logging.set_verbosity(logging.INFO)
    app.run(_main, argv=["pseudo.py"])



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import re

import tensorflow as tf
from absl import logging

from neurst.utils.checkpoints import NameBasedCheckpointManager

GLOBAL_SETTING = dict()
FLOAT_MIN = -1.e9
CUSTOM_GLOBAL_FLOATX = "float32"


def _broadcast_global_setting(name, var):
    global GLOBAL_SETTING
    GLOBAL_SETTING[name] = var


def _get_global_setting_or_default(name, default=None):
    global GLOBAL_SETTING
    return GLOBAL_SETTING.get(name, default)


class ModeKeys(object):
    """ Standard names for model modes. """
    TRAIN = "train"
    INFER = "infer"
    EVAL = "eval"

    _fields = ["train", "infer", "eval"]


class DataStatus(object):
    """ Standard names for data sample status. """
    # the status of the data,
    # where RAW - the data is unprocessed,
    #       PROCESSED - the data is already processed,
    #       PROJECTED - the data is already projected into network readable format (e.g. integer)
    RAW = "raw"
    PROCESSED = "processed"
    PROJECTED = "projected"


class GlobalKeys:
    # Key to indicate various ops.
    INITIAL_GLOBAL_STEP = "global_step"
    SUMMARY_STEPS = "summary_steps"
    KERAS_TRAIN_MODEL = "train_model"

    # computation precision
    FLOAT_DTYPE = "FLOAT_DTYPE"
    FLOAT_MIN = "FLOAT_MIN"

    # Metrics prefix for tensorboard
    TBPREFIX_TRAINING = "training"
    TBPREFIX_VALID = "valid"

    # checkpoint manager
    SAVER = "checkpoint_manager"

    # distributed training workers
    DIST_WORKER_ID = "distributed_worker_id"
    DIST_NUM_WORKERS = "distributed_num_workers"
    DIST_STRATEGY = "distributed_strategy"


def register_initial_step(step):
    """ Memorizes the initial step. """
    _broadcast_global_setting(GlobalKeys.INITIAL_GLOBAL_STEP, step)


def get_registered_initial_step():
    """ Gets the registered initial step. """
    return _get_global_setting_or_default(GlobalKeys.INITIAL_GLOBAL_STEP, 0)


def register_distributed_worker_setting(worker_id, num_workers, strategy):
    """ Memorizes the current worker id and total number of workers. """
    logging.info(f"Register distribution strategy: {str(strategy)} "
                 f"on worker {worker_id} (total {num_workers}).")
    _broadcast_global_setting(GlobalKeys.DIST_WORKER_ID, worker_id)
    _broadcast_global_setting(GlobalKeys.DIST_NUM_WORKERS, num_workers)
    _broadcast_global_setting(GlobalKeys.DIST_STRATEGY, strategy)


def get_distributed_worker_setting():
    return (_get_global_setting_or_default(GlobalKeys.DIST_WORKER_ID, 0),
            _get_global_setting_or_default(GlobalKeys.DIST_NUM_WORKERS, 1),
            _get_global_setting_or_default(GlobalKeys.DIST_STRATEGY, None))


def register_computation_dtype(floatx, float_min):
    """ Register computation type of this program. """
    global CUSTOM_GLOBAL_FLOATX
    global FLOAT_MIN
    CUSTOM_GLOBAL_FLOATX = floatx
    FLOAT_MIN = float_min


def get_saver_or_default(model=None, model_dir=None, **kwargs):
    """ Creates the default checkpoint manager.

    Args:
        model: A tf.keras.model.Model (should be already called once).
        model_dir: The save to path, either on disk or HDFS.
        kwargs: Arbitrary augments for initializing custom checkpoint manager.
    Returns: A checkpoint manager
    """
    checkpoint_manager = _get_global_setting_or_default(
        GlobalKeys.SAVER, default=None)
    if checkpoint_manager is None:
        assert model is not None, (
            "`model` must be provided for default saver creation.")
        assert model_dir, (
            "`model_dir` must be provided for default saver creation.")
        checkpoint_manager = NameBasedCheckpointManager(
            model=model, directory=model_dir, **kwargs)
        _broadcast_global_setting(GlobalKeys.SAVER, checkpoint_manager)
    assert checkpoint_manager is not None
    return checkpoint_manager


def hack_global_step(checkpoint_dir):
    """ Returns the global step according to the checkpoint name. """
    if tf.io.gfile.isdir(checkpoint_dir):
        checkpoint_dir = tf.train.latest_checkpoint(checkpoint_dir)
    if checkpoint_dir is None:
        return None
    base_name = os.path.basename(checkpoint_dir)
    re_search = re.search(r"\d{2,}", base_name)
    if re_search:
        return int(re_search.group())
    return None


def wrapper_var_name(name_from_ckpt):
    name = name_from_ckpt.replace(".S", "/")
    name = name.replace("/.ATTRIBUTES/VARIABLE_VALUE", "")
    return name



# coding: utf-8
from __future__ import absolute_import, division, print_function, unicode_literals

import importlib

from .__version__ import __version__  # NOQA

__author__ = "ZhaoChengqi <zhaochengqi.d@bytedance.com>"

__all__ = [
    "cli",
    "data",
    "criterions",
    "exps",
    "layers",
    "metrics",
    "models",
    "tasks",
    "utils",
    "training",
    "optimizers"
]

importlib.import_module("neurst.criterions")
importlib.import_module("neurst.data")
importlib.import_module("neurst.data.audio")
importlib.import_module("neurst.data.data_pipelines")
importlib.import_module("neurst.data.datasets")
importlib.import_module("neurst.data.datasets.audio")
importlib.import_module("neurst.data.text")
importlib.import_module("neurst.exps")
importlib.import_module("neurst.layers")
importlib.import_module("neurst.layers.attentions")
importlib.import_module("neurst.layers.decoders")
importlib.import_module("neurst.layers.encoders")
importlib.import_module("neurst.layers.metric_layers")
importlib.import_module("neurst.layers.search")
importlib.import_module("neurst.metrics")
importlib.import_module("neurst.models")
importlib.import_module("neurst.optimizers")
importlib.import_module("neurst.optimizers.schedules")
importlib.import_module("neurst.tasks")
importlib.import_module("neurst.training")
importlib.import_module("neurst.utils")
importlib.import_module("neurst.utils.converters")



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import tensorflow as tf

from neurst.layers.metric_layers import METRIC_REDUCTION
from neurst.layers.metric_layers.metric_layer import MetricLayer


class SequenceTokenMetricLayer(MetricLayer):
    """Custom a layer of metrics for Transformer model."""

    def __init__(self, name_prefix):
        super(SequenceTokenMetricLayer, self).__init__()
        self._name_prefix = name_prefix

    def build(self, input_shape):
        super(SequenceTokenMetricLayer, self).build(input_shape)
        self.build_metric_reduction(self._name_prefix + "_tokens", METRIC_REDUCTION.SUM)
        self.build_metric_reduction(self._name_prefix + "_real_tokens", METRIC_REDUCTION.SUM)

    def calculate(self, input, output):
        """ Calculates metric values according to model input and output. """
        x = input[self._name_prefix]
        ms = {self._name_prefix + "_tokens": tf.cast(tf.shape(x)[0] * tf.shape(x)[1], tf.float32)}
        if self._name_prefix + "_padding" in input:
            x_len = ms[self._name_prefix + "_tokens"] - tf.cast(
                tf.reduce_sum(input[self._name_prefix + "_padding"]), tf.float32)
        else:
            x_len = tf.reduce_sum(tf.cast(input[self._name_prefix + "_length"], tf.float32))
        ms[self._name_prefix + "_real_tokens"] = x_len
        return ms


class AudioFramesMetricLayer(MetricLayer):
    """Custom a layer of metrics for Speech Transformer model."""

    def __init__(self, name_prefix):
        super(AudioFramesMetricLayer, self).__init__()
        self._name_prefix = name_prefix

    def build(self, input_shape):
        """" Builds metric layer. """
        super(AudioFramesMetricLayer, self).build(input_shape)
        self.build_metric_reduction(self._name_prefix + "_tokens", METRIC_REDUCTION.SUM)
        self.build_metric_reduction(self._name_prefix + "_real_tokens", METRIC_REDUCTION.SUM)

    def calculate(self, input, output):
        """ Calculates metric values according to model input and output. """
        x = input[self._name_prefix]
        x_len = input[self._name_prefix + "_length"]
        ms = {self._name_prefix + "_tokens": tf.cast(tf.shape(x)[0] * tf.shape(x)[1], tf.float32),
              self._name_prefix + "_real_tokens": tf.cast(tf.reduce_sum(x_len), tf.float32)}
        return ms


class BatchCountMetricLayer(MetricLayer):
    """Custom a layer of metrics for Transformer model."""

    def __init__(self, key):
        super(BatchCountMetricLayer, self).__init__()
        self._key = key

    def build(self, input_shape):
        super(BatchCountMetricLayer, self).build(input_shape)
        self.build_metric_reduction("samples", METRIC_REDUCTION.SUM)

    def calculate(self, input, output):
        """ Calculates metric values according to model input and output. """
        x = input[self._key]
        return {"samples": tf.cast(tf.shape(x)[0], tf.float32)}



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import copy
import re
import traceback

import tensorflow as tf
from absl import logging
from tensorflow.keras import backend as K
from tensorflow.python.distribute import parameter_server_strategy
from tensorflow.python.eager import backprop, context
from tensorflow.python.framework import ops
from tensorflow.python.keras import callbacks as callbacks_module
from tensorflow.python.keras.engine import data_adapter, training, training_utils
from tensorflow.python.keras.mixed_precision.experimental import loss_scale_optimizer as lso
from tensorflow.python.keras.utils import version_utils
from tensorflow.python.profiler import trace


def _multiply_gradient(gradient, scale):
    """Multiply a (possibly sparse) gradient by the given scale factor."""
    if gradient is None:
        return None
    if isinstance(gradient, ops.IndexedSlices):
        return ops.IndexedSlices(
            gradient.values * tf.convert_to_tensor(scale, dtype=gradient.dtype),
            gradient.indices,
            dense_shape=gradient.dense_shape)
    else:
        return gradient * scale


def _minimum_control_deps(outputs):
    """Returns the minimum control dependencies to ensure step succeeded."""
    if context.executing_eagerly():
        return []  # Control dependencies not needed.
    outputs = tf.nest.flatten(outputs, expand_composites=True)
    for out in outputs:
        # Variables can't be control dependencies.
        if not isinstance(out, tf.Variable):
            return [out]  # Return first Tensor or Op from outputs.
    return []  # No viable Tensor or Op to use for control deps.


class GradientAccumulator(object):
    """Gradient accumulation utility.
        When used with a distribution strategy, the accumulator should be called in a
        replica context. Gradients will be accumulated locally on each replica and
        without synchronization. Users should then call ``.gradients``, scale the
        gradients if required, and pass the result to ``apply_gradients``.
        """

    # We use the ON_READ synchronization policy so that no synchronization is
    # performed on assignment. To get the value, we call .value() which returns the
    # value on the current replica without synchronization.
    def __init__(self, accum_steps):
        """Initializes the accumulator."""
        self._gradients = []
        self._accum_grad_scale = 1. / float(accum_steps)

    @property
    def gradients(self):
        """The accumulated gradients on the current replica."""
        return [gradient.value() if gradient is not None else gradient
                for gradient in self._gradients]

    def __call__(self, gradients):
        """Accumulates :obj:`gradients` on the current replica."""
        if len(self._gradients) == 0:
            self._gradients.extend([
                tf.Variable(
                    tf.zeros_like(gradient),
                    trainable=False,
                    synchronization=tf.VariableSynchronization.ON_READ,
                    aggregation=tf.VariableAggregation.ONLY_FIRST_REPLICA)
                if gradient is not None else gradient for gradient in gradients])

        for accum_gradient, gradient in zip(self._gradients, gradients):
            if accum_gradient is not None and gradient is not None:
                K.update_add(accum_gradient, _multiply_gradient(gradient, self._accum_grad_scale))

    def reset(self):
        """Resets the accumulated gradients on the current replica."""
        for gradient in self._gradients:
            if gradient is not None:
                K.update(gradient, tf.zeros_like(gradient))


class AccumgradKerasModel(tf.keras.Model):
    """ Defines the keras model that supports gradient accumulation. """

    def __init__(self, *args, **kwargs):
        self._update_cycle = int(kwargs.pop("update_cycle", 1))
        if not self._update_cycle:
            self._update_cycle = 1
        self._clip_value = kwargs.pop("clip_value", None)
        self._clip_norm = kwargs.pop("clip_norm", None)
        self._freeze_variables = kwargs.pop("freeze_variables", None)
        super(AccumgradKerasModel, self).__init__(*args, **kwargs)
        if self._clip_value:
            self._clip_value = abs(float(self._clip_value))
            logging.info(f"Clipping gradient to = [-{self._clip_value}, {self._clip_value}]")
        elif self._clip_norm:
            logging.info(f"Clipping gradient norm to {self._clip_norm}")
        if self._update_cycle > 1:
            logging.info(f"Accumulating gradients for every {self._update_cycle} steps.")
            self._grad_accumulator = GradientAccumulator(self._update_cycle)
        if self._freeze_variables:
            logging.info(f"Variable names matched the pattern {self._freeze_variables} will be freezed.")
        self.accumulate_function = None

    @property
    def custom_trainable_weights(self):
        if self._freeze_variables:
            return [x for x in self.trainable_weights
                    if re.search(self._freeze_variables, x.name) is None]
        return self.trainable_weights

    @property
    def custom_trainable_variables(self):
        return self.custom_trainable_weights

    def compute_gradients(self, data):
        """The logic for gradient computation of one training step.

        This method can be overridden to support custom training logic.
        This method is called by `Model.make_train_function`.

        This method should contain the mathemetical logic for one step of training.
        This typically includes the forward pass, loss calculation, backpropagation,
        and metric updates.

        Configuration details for *how* this logic is run (e.g. `tf.function` and
        `tf.distribute.Strategy` settings), should be left to
        `Model.make_train_function`, which can also be overridden.

        Arguments:
          data: A nested structure of `Tensor`s.

        Returns:
          A list of gradients corresponding to `Model.trainable_variables`.

        """
        # These are the only transformations `Model.fit` applies to user-input
        # data when a `tf.data.Dataset` is provided. These utilities will be exposed
        # publicly.
        data = data_adapter.expand_1d(data)
        x, y, sample_weight = data_adapter.unpack_x_y_sample_weight(data)

        with backprop.GradientTape() as tape:
            y_pred = self(x, training=True)
            loss = self.compiled_loss(
                y, y_pred, sample_weight, regularization_losses=self.losses)

        # For custom training steps, users can just write:
        #   trainable_variables = self.trainable_variables
        #   gradients = tape.gradient(loss, trainable_variables)
        #   self.optimizer.apply_gradients(zip(gradients, trainable_variables))
        # The _minimize call does a few extra steps unnecessary in most cases,
        # such as loss scaling and gradient clipping.
        with tape:
            if isinstance(self.optimizer, lso.LossScaleOptimizer):
                loss = self.optimizer.get_scaled_loss(loss)
            gradients = tape.gradient(loss, self.custom_trainable_variables)

        def _zeros_grads():
            zero_gradients = []
            for gradient in gradients:
                if isinstance(gradient, ops.IndexedSlices):
                    zero_gradients.append(ops.IndexedSlices(
                        tf.zeros_like(gradient.values),
                        gradient.indices,
                        dense_shape=gradient.dense_shape))
                elif gradient is None:
                    zero_gradients.append(None)
                else:
                    zero_gradients.append(tf.zeros_like(gradient))
            return zero_gradients

        gradients = tf.cond(tf.math.reduce_all(tf.math.is_finite(loss)),
                            lambda: gradients, _zeros_grads)

        self.compiled_metrics.update_state(y, y_pred, sample_weight)
        return gradients, {m.name: m.result() for m in self.metrics}

    def custom_apply_gradients(self, gradients):
        trainable_variables = self.custom_trainable_variables
        # Whether to aggregate gradients outside of optimizer. This requires support
        # of the optimizer and doesn't work with ParameterServerStrategy and
        # CentralStroageStrategy.
        aggregate_grads_outside_optimizer = (
            self.optimizer._HAS_AGGREGATE_GRAD  # pylint: disable=protected-access
            and not isinstance(self.distribute_strategy.extended,
                               parameter_server_strategy.ParameterServerStrategyExtended))
        if aggregate_grads_outside_optimizer:
            # We aggregate gradients before unscaling them, in case a subclass of
            # LossScaleOptimizer all-reduces in fp16. All-reducing in fp16 can only be
            # done on scaled gradients, not unscaled gradients, for numeric stability.
            gradients = self.optimizer._aggregate_gradients(zip(gradients,  # pylint: disable=protected-access
                                                                trainable_variables))
        if isinstance(self.optimizer, lso.LossScaleOptimizer):
            gradients = self.optimizer.get_unscaled_gradients(gradients)
        # gradients = self.optimizer._clip_gradients(gradients)  # pylint: disable=protected-access
        if self._clip_value:
            gradients = [tf.clip_by_value(grad, -self._clip_value, self._clip_value)
                         if grad is not None else None for grad in gradients]
        elif self._clip_norm:
            gradients = [tf.clip_by_norm(grad, self._clip_norm)
                         if grad is not None else None for grad in gradients]
        if trainable_variables:
            if aggregate_grads_outside_optimizer:
                self.optimizer.apply_gradients(zip(gradients, trainable_variables),
                                               experimental_aggregate_gradients=False)
            else:
                self.optimizer.apply_gradients(zip(gradients, trainable_variables))

    def make_train_function(self):
        """Creates a function that executes one step of training.

            This method can be overridden to support custom training logic.
            This method is called by `Model.fit` and `Model.train_on_batch`.

            Typically, this method directly controls `tf.function` and
            `tf.distribute.Strategy` settings, and delegates the actual training
            logic to `Model.train_step`.

            This function is cached the first time `Model.fit` or
            `Model.train_on_batch` is called. The cache is cleared whenever
            `Model.compile` is called.

            Returns:
              Function. The function created by this method should accept a
              `tf.data.Iterator`, and return a `dict` containing values that will
              be passed to `tf.keras.Callbacks.on_train_batch_end`, such as
              `{'loss': 0.2, 'accuracy': 0.7}`.
            """
        if self.train_function is not None:
            return self.train_function

        def step_function(model, iterator):
            """Runs a single training step."""

            def run_step(data):
                gradients, outputs = model.compute_gradients(data)
                if self._update_cycle > 1:
                    model._grad_accumulator(gradients)  # pylint: disable=protected-access
                    gradients = model._grad_accumulator.gradients  # pylint: disable=protected-access
                model.custom_apply_gradients(gradients)
                if model._update_cycle > 1:  # pylint: disable=protected-access
                    model._grad_accumulator.reset()  # pylint: disable=protected-access
                # Ensure counter is updated only if `train_step` succeeds.
                with ops.control_dependencies(_minimum_control_deps(outputs)):
                    model._train_counter.assign_add(1)  # pylint: disable=protected-access
                return outputs

            err_cnt = 0
            while err_cnt < 10:
                try:
                    data = next(iterator)
                    break
                except (StopIteration, tf.errors.OutOfRangeError) as e:
                    raise e
                except tf.errors.OpError as e:
                    tf.print(traceback.format_exc(e))
                    err_cnt += 1
                    if err_cnt >= 10:
                        raise e

            outputs = model.distribute_strategy.run(run_step, args=(data,))
            outputs = training.reduce_per_replica(outputs, self.distribute_strategy)
            training.write_scalar_summaries(outputs, step=model._train_counter)  # pylint: disable=protected-access
            return outputs

        def step_function_accum(model, iterator):
            """Runs a single training step."""

            def run_step(data):
                gradients, outputs = model.compute_gradients(data)
                model._grad_accumulator(gradients)  # pylint: disable=protected-access

            err_cnt = 0
            while err_cnt < 10:
                try:
                    data = next(iterator)
                    break
                except (StopIteration, tf.errors.OutOfRangeError) as e:
                    raise e
                except tf.errors.OpError as e:
                    tf.print(traceback.format_exc(e))
                    err_cnt += 1
                    if err_cnt >= 10:
                        raise e

            model.distribute_strategy.run(run_step, args=(data,))

        def accumulate_function(iterator):
            return step_function_accum(self, iterator)

        def train_function(iterator):
            """Runs a training execution with one step."""
            return step_function(self, iterator)

        if not self.run_eagerly:
            train_function = tf.function(
                train_function, experimental_relax_shapes=True)
            if self._update_cycle > 1:
                accumulate_function = tf.function(
                    accumulate_function, experimental_relax_shapes=True)

        self.train_function = train_function
        if self._update_cycle > 1:
            self.accumulate_function = accumulate_function
        return self.train_function

    @training.enable_multi_worker
    def fit(self,
            x=None,
            y=None,
            batch_size=None,
            epochs=1,
            verbose=1,
            callbacks=None,
            validation_split=0.,
            validation_data=None,
            shuffle=True,
            class_weight=None,
            sample_weight=None,
            initial_epoch=0,
            steps_per_epoch=None,
            validation_steps=None,
            validation_batch_size=None,
            validation_freq=1,
            max_queue_size=10,
            workers=1,
            use_multiprocessing=False):
        """ Copy from tf.keras.Model. """
        training._keras_api_gauge.get_cell('fit').set(True)
        # Legacy graph support is contained in `training_v1.Model`.
        version_utils.disallow_legacy_graph('Model', 'fit')
        self._assert_compile_was_called()
        self._check_call_args('fit')
        training._disallow_inside_tf_function('fit')

        if validation_split:
            # Create the validation data using the training data. Only supported for
            # `Tensor` and `NumPy` input.
            (x, y, sample_weight), validation_data = (
                data_adapter.train_validation_split(
                    (x, y, sample_weight), validation_split=validation_split))

        if validation_data:
            val_x, val_y, val_sample_weight = (
                data_adapter.unpack_x_y_sample_weight(validation_data))

        with self.distribute_strategy.scope(), \
             training_utils.RespectCompiledTrainableState(self):
            # Creates a `tf.data.Dataset` and handles batch and epoch iteration.
            data_handler = data_adapter.DataHandler(
                x=x,
                y=y,
                sample_weight=sample_weight,
                batch_size=batch_size,
                steps_per_epoch=steps_per_epoch,
                initial_epoch=initial_epoch,
                epochs=epochs,
                shuffle=shuffle,
                class_weight=class_weight,
                max_queue_size=max_queue_size,
                workers=workers,
                use_multiprocessing=use_multiprocessing,
                model=self,
                steps_per_execution=self._steps_per_execution)

            # Container that configures and calls `tf.keras.Callback`s.
            if not isinstance(callbacks, callbacks_module.CallbackList):
                callbacks = callbacks_module.CallbackList(
                    callbacks,
                    add_history=True,
                    add_progbar=verbose != 0,
                    model=self,
                    verbose=verbose,
                    epochs=epochs,
                    steps=data_handler.inferred_steps)

            self.stop_training = False
            train_function = self.make_train_function()
            self._train_counter.assign(0)
            callbacks.on_train_begin()
            training_logs = None
            # Handle fault-tolerance for multi-worker.
            # TODO(omalleyt): Fix the ordering issues that mean this has to
            # happen after `callbacks.on_train_begin`.
            data_handler._initial_epoch = (  # pylint: disable=protected-access
                self._maybe_load_initial_epoch_from_ckpt(initial_epoch))
            for epoch, iterator in data_handler.enumerate_epochs():
                self.reset_metrics()
                callbacks.on_epoch_begin(epoch)
                with data_handler.catch_stop_iteration():
                    if self._update_cycle > 1:
                        self._grad_accumulator.reset()
                    for step in data_handler.steps():
                        with trace.Trace(
                            'TraceContext',
                            graph_type='train',
                            epoch_num=epoch,
                            step_num=step,
                            batch_size=batch_size):
                            callbacks.on_train_batch_begin(step)
                            if self._update_cycle > 1:
                                for _ in range(self._update_cycle - 1):
                                    self.accumulate_function(iterator)
                            tmp_logs = train_function(iterator)
                            if data_handler.should_sync:
                                context.async_wait()
                            logs = tmp_logs  # No error, now safe to assign to logs.
                            end_step = step + data_handler.step_increment
                            callbacks.on_train_batch_end(end_step, logs)
                epoch_logs = copy.copy(logs)

                # Run validation.
                if validation_data and self._should_eval(epoch, validation_freq):
                    # Create data_handler for evaluation and cache it.
                    if getattr(self, '_eval_data_handler', None) is None:
                        self._eval_data_handler = data_adapter.DataHandler(
                            x=val_x,
                            y=val_y,
                            sample_weight=val_sample_weight,
                            batch_size=validation_batch_size or batch_size,
                            steps_per_epoch=validation_steps,
                            initial_epoch=0,
                            epochs=1,
                            max_queue_size=max_queue_size,
                            workers=workers,
                            use_multiprocessing=use_multiprocessing,
                            model=self,
                            steps_per_execution=self._steps_per_execution)
                    val_logs = self.evaluate(
                        x=val_x,
                        y=val_y,
                        sample_weight=val_sample_weight,
                        batch_size=validation_batch_size or batch_size,
                        steps=validation_steps,
                        callbacks=callbacks,
                        max_queue_size=max_queue_size,
                        workers=workers,
                        use_multiprocessing=use_multiprocessing,
                        return_dict=True)
                    val_logs = {'val_' + name: val for name, val in val_logs.items()}
                    epoch_logs.update(val_logs)

                callbacks.on_epoch_end(epoch, epoch_logs)
                training_logs = epoch_logs
                if self.stop_training:
                    break

            # If eval data_hanlder exists, delete it after all epochs are done.
            if getattr(self, '_eval_data_handler', None) is not None:
                del self._eval_data_handler
            callbacks.on_train_end(logs=training_logs)
            return self.history



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import numpy
import tensorflow as tf

from neurst.layers.common_layers import PositionEmbeddingWrapper
from neurst.layers.decoders import build_decoder
from neurst.layers.encoders import build_encoder
from neurst.layers.modalities.audio_modalities import AudioConvSubsamplingLayer
from neurst.models import register_model
from neurst.models.encoder_decoder_model import EncoderDecoderModel
from neurst.utils import compat
from neurst.utils.flags_core import Flag
from neurst.utils.hparams_sets import register_hparams_set


@register_model
class SpeechTransformer(EncoderDecoderModel):
    """ Defines the Speech Transformer model. """

    def __init__(self, args, *largs, **kwargs):
        super(SpeechTransformer, self).__init__(args, *largs, **kwargs)
        self._args = args

    @staticmethod
    def class_or_method_args():
        return [
            Flag("modality.share_embedding_and_softmax_weights", dtype=Flag.TYPE.BOOLEAN, default=False,
                 help="Whether to share the target embedding table and softmax weights."),
            Flag("modality.dim", dtype=Flag.TYPE.INTEGER, default=None,
                 help="The default embedding dimension for both source and target side."),
            Flag("modality.source.dim", dtype=Flag.TYPE.INTEGER, default=None,
                 help="The source-side embedding dimension, or `modality.dim` if not provided."),
            Flag("modality.target.dim", dtype=Flag.TYPE.INTEGER, default=None,
                 help="The target-side embedding dimension, or `modality.dim` if not provided."),
            Flag("modality.timing", dtype=Flag.TYPE.STRING, default=None,
                 help="The arbitrary parameters for positional encoding of both source and target side."),
            Flag("modality.source.timing", dtype=Flag.TYPE.STRING, default=None,
                 help="The arbitrary parameters for source-side positional encoding, "
                      "or `modality.timing` if not provided."),
            Flag("modality.target.timing", dtype=Flag.TYPE.STRING, default=None,
                 help="The arbitrary parameters for target-side positional encoding, "
                      "or `modality.timing` if not provided."),
            Flag("modality.source.kernel_size", dtype=Flag.TYPE.INTEGER, default=3,
                 help="The kernel size for the first two conv layer"),
            Flag("modality.source.strides", dtype=Flag.TYPE.INTEGER, default=2,
                 help="The stride size for the first two conv layer"),
            Flag("modality.source.channels", dtype=Flag.TYPE.INTEGER, default=256,
                 help="The channels for the first two conv layer"),
            Flag("modality.source.layer_norm", dtype=Flag.TYPE.BOOLEAN, default=True,
                 help="Whether to apply layer norm in convolution layers."),
            Flag("encoder.num_layers", dtype=Flag.TYPE.INTEGER, default=None,
                 help="The number of stacking layers of the encoder."),
            Flag("encoder.hidden_size", dtype=Flag.TYPE.INTEGER, default=None,
                 help="The number of hidden units of the encoder."),
            Flag("encoder.num_attention_heads", dtype=Flag.TYPE.INTEGER, default=None,
                 help="The number of heads of encoder self-attention."),
            Flag("encoder.filter_size", dtype=Flag.TYPE.INTEGER, default=None,
                 help="The number of the filter size of encoder ffn."),
            Flag("encoder.ffn_activation", dtype=Flag.TYPE.STRING, default="relu",
                 help="The activation function of encoder ffn layer."),
            Flag("encoder.attention_dropout_rate", dtype=Flag.TYPE.FLOAT, default=0.,
                 help="The dropout rate of encoder self-attention layer."),
            Flag("encoder.attention_type", dtype=Flag.TYPE.STRING, default="dot_product",
                 help="The type of the attention function of encoder self-attention layer."),
            Flag("encoder.ffn_dropout_rate", dtype=Flag.TYPE.FLOAT, default=0.,
                 help="The dropout rate of encoder ffn layer."),
            Flag("encoder.layer_postprocess_dropout_rate", dtype=Flag.TYPE.FLOAT, default=0.,
                 help="The dropout rate for each layer's post process in encoder."),
            Flag("encoder.layer_postprocess_epsilon", dtype=Flag.TYPE.FLOAT, default=1e-6,
                 help="The epsilon for layer normalization in decoder."),
            Flag("decoder.num_layers", dtype=Flag.TYPE.INTEGER, default=None,
                 help="The number of stacking layers of the decoder."),
            Flag("decoder.hidden_size", dtype=Flag.TYPE.INTEGER, default=None,
                 help="The number of hidden units of the decoder."),
            Flag("decoder.num_attention_heads", dtype=Flag.TYPE.INTEGER, default=None,
                 help="The number of heads of decoder self-attention and encoder-decoder attention."),
            Flag("decoder.filter_size", dtype=Flag.TYPE.INTEGER, default=None,
                 help="The number of the filter size of decoder ffn."),
            Flag("decoder.ffn_activation", dtype=Flag.TYPE.STRING, default="relu",
                 help="The activation function of decoder ffn layer."),
            Flag("decoder.attention_dropout_rate", dtype=Flag.TYPE.FLOAT, default=0.,
                 help="The dropout rate of decoder self-attention and encoder-decoder attention."),
            Flag("decoder.attention_type", dtype=Flag.TYPE.STRING, default="dot_product",
                 help="The type of the attention function of decoder self-attention and encoder-decoder attention."),
            Flag("decoder.ffn_dropout_rate", dtype=Flag.TYPE.FLOAT, default=0.,
                 help="The dropout rate of decoder ffn layer."),
            Flag("decoder.layer_postprocess_dropout_rate", dtype=Flag.TYPE.FLOAT, default=0.,
                 help="The dropout rate for each layer's post process in decoder."),
            Flag("decoder.layer_postprocess_epsilon", dtype=Flag.TYPE.FLOAT, default=1e-6,
                 help="The epsilon for layer normalization in decoder."),
        ]

    @classmethod
    def build_modalities(cls, model_args, src_meta, trg_meta):
        """ Create source and target modality. """
        # if modality.source.dim is not defined, then use modality.dim as default
        src_dim = model_args["modality.source.dim"] or model_args["modality.dim"]
        # if modality.target.dim is not defined, then use modality.dim as default
        trg_dim = model_args["modality.target.dim"] or model_args["modality.dim"]
        # whether to share source and target embedding
        input_name = "input_audio_modality"
        target_name = "target_symbol_modality"
        # creates target embedding table
        target_modality = cls.build_modality(
            vocab_size=trg_meta["vocab_size"], emb_dim=trg_dim, name=target_name,
            timing=(model_args["modality.target.timing"] or model_args["modality.timing"]),
            share_embedding_and_softmax_weights=model_args["modality.share_embedding_and_softmax_weights"])

        # creates source audio modality
        input_modality = AudioConvSubsamplingLayer(
            embedding_dim=src_dim,
            kernel_size=model_args["modality.source.kernel_size"],
            strides=model_args["modality.source.strides"],
            channels=model_args["modality.source.channels"],
            layer_norm=model_args["modality.source.layer_norm"],
            name=input_name)
        src_timing = model_args["modality.source.timing"] or model_args["modality.timing"]
        if src_timing:
            if isinstance(src_timing, str):
                src_timing = {"timing": src_timing}
            elif not isinstance(src_timing, dict):
                raise ValueError("Unknown type of timing params: {}".format(str(src_timing)))
            input_modality = PositionEmbeddingWrapper(
                embedding_layer=input_modality, name=input_name + "_posenc_wrapper", **src_timing)
        return input_modality, target_modality

    @classmethod
    def new(cls, args, src_meta, trg_meta, name=None):
        """ Builds a sequence to sequence model.

        Args:
            args: A dict containing all model parameters.
            src_meta: A dict containing source-side vocabulary meta data, e.g. eos_id, vocab_size.
            trg_meta: A dict containing target-side vocabulary meta data, e.g. eos_id, vocab_size.
            name: The name of the model.

        Returns:
            An encoder decoder model.
        """
        # build source and target modality
        src_modality, trg_modality = cls.build_modalities(args, src_meta, trg_meta)
        encoder_params, decoder_params = {}, {}
        for f in cls.class_or_method_args():
            if f.name in args:
                if f.name.startswith("encoder."):
                    encoder_params[f.name[8:]] = args[f.name]
                elif f.name.startswith("decoder."):
                    decoder_params[f.name[8:]] = args[f.name]
        # build encoder and decoder
        encoder = build_encoder({
            "encoder.class": "TransformerEncoder",
            "encoder.params": encoder_params})
        decoder = build_decoder({
            "decoder.class": "TransformerDecoder",
            "decoder.params": decoder_params})
        model = cls(args, src_meta, trg_meta, src_modality, trg_modality, encoder, decoder, name=name)
        fake_src = numpy.random.rand(1, 4, src_meta["audio_feature_dim"], src_meta["audio_feature_channels"])
        fake_inputs = {"src": tf.convert_to_tensor(fake_src, tf.float32),
                       "src_length": tf.convert_to_tensor([4], tf.int64),
                       "trg_input": tf.convert_to_tensor([[1, 2, 3]], tf.int64), }
        _ = model(fake_inputs)
        return model

    def get_symbols_to_logits_fn(self, inputs, *args, **kwargs):
        strides = self.args["modality.source.strides"]

        def _length_after_conv(_l):
            return ((_l + strides - 1) // strides + strides - 1) // strides

        inputs["src_padding"] = 1. - tf.sequence_mask(
            lengths=tf.cast(_length_after_conv(inputs["src_length"]), tf.int32),
            maxlen=tf.cast(_length_after_conv(tf.shape(inputs["src"])[1]), tf.int32),
            dtype=tf.dtypes.as_dtype(compat.CUSTOM_GLOBAL_FLOATX))
        return super(SpeechTransformer, self).get_symbols_to_logits_fn(inputs, *args, **kwargs)

    @classmethod
    def build_model_args_by_name(cls, name):
        if not name.startswith("speech_transformer"):
            return None

        kernel_size = 3
        strides = 2
        channels = 256
        layer_norm = True
        if name == "speech_transformer_toy":
            dmodel = 8
            num_heads = 2
            num_encoder_layers = 2
            num_decoder_layers = 2
            num_encoder_filter_size = 10
            num_decoder_filter_size = 10
            channels = 5
            dropout_rate = 0.1
        elif name == "speech_transformer_m":
            dmodel = 512
            num_heads = 8
            num_encoder_layers = 12
            num_decoder_layers = 6
            num_encoder_filter_size = 2048
            num_decoder_filter_size = 2048
            dropout_rate = 0.1
        elif name == "speech_transformer_s":
            dmodel = 256
            num_heads = 4
            num_encoder_layers = 12
            num_decoder_layers = 6
            num_encoder_filter_size = 2048
            num_decoder_filter_size = 2048
            dropout_rate = 0.1
        else:
            return None
        return {
            "model.class": cls.__name__,
            "model.params": {
                "modality.source.kernel_size": kernel_size,
                "modality.source.strides": strides,
                "modality.source.channels": channels,
                "modality.source.layer_norm": layer_norm,
                "modality.dim": dmodel,
                "modality.share_embedding_and_softmax_weights": True,
                "modality.timing": "sinusoids",
                "encoder.num_layers": num_encoder_layers,
                "encoder.hidden_size": dmodel,
                "encoder.num_attention_heads": num_heads,
                "encoder.filter_size": num_encoder_filter_size,
                "encoder.attention_dropout_rate": dropout_rate,
                "encoder.attention_type": "dot_product",
                "encoder.ffn_activation": "relu",
                "encoder.ffn_dropout_rate": dropout_rate,
                "encoder.layer_postprocess_dropout_rate": dropout_rate,
                "decoder.num_layers": num_decoder_layers,
                "decoder.hidden_size": dmodel,
                "decoder.num_attention_heads": num_heads,
                "decoder.filter_size": num_decoder_filter_size,
                "decoder.attention_dropout_rate": dropout_rate,
                "decoder.attention_type": "dot_product",
                "decoder.ffn_activation": "relu",
                "decoder.ffn_dropout_rate": dropout_rate,
                "decoder.layer_postprocess_dropout_rate": dropout_rate
            },
            "optimizer.class": "Adam",
            "optimizer.params": {
                "epsilon": 1.e-9,
                "beta_1": 0.9,
                "beta_2": 0.98
            },
            "lr_schedule.class": "noam",
            "lr_schedule.params": {
                "initial_factor": 5.0 if dmodel > 256 else 3.5,
                "end_factor": 2.0 if dmodel > 256 else 1.5,
                "dmodel": dmodel,
                "warmup_steps": 25000,
                "start_decay_at": 50000,
                "decay_steps": 50000,
            },
        }


@register_hparams_set("speech_transformer_toy")
def speech_transformer_toy():
    return SpeechTransformer.build_model_args_by_name("speech_transformer_toy")


@register_hparams_set("speech_transformer_base")
def speech_transformer_base():
    return SpeechTransformer.build_model_args_by_name("speech_transformer_base")



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from abc import ABCMeta, abstractmethod

import six


@six.add_metaclass(ABCMeta)
class Metric(object):
    REGISTRY_NAME = "metric"

    def __init__(self, *args, **kwargs):
        self._flag = self.__class__.__name__

    @property
    def flag(self):
        return self._flag

    @flag.setter
    def flag(self, flag_name):
        """ Sets the flag metric name if the result of `__call__` is a dict. """
        self._flag = flag_name

    def set_groundtruth(self, groundtruth):
        raise NotImplementedError

    def greater_or_eq(self, result1, result2):
        """ Compare the two metric value result and return True if v1>=v2. """
        return self.get_value(result1) >= self.get_value(result2)

    def get_value(self, result):
        """ Gets a float value from the metric result (if is a dict). """
        if isinstance(result, dict) and self._flag in result:
            return result[self._flag]
        return result

    def __call__(self, hypothesis, groundtruth=None) -> dict:
        """ Returns a dict of metric values. """
        res = self.call(hypothesis, groundtruth=groundtruth)
        if not isinstance(res, dict):
            res = {self.flag: res}
        return res

    @abstractmethod
    def call(self, hypothesis, groundtruth=None):
        """ Returns the metric value (float) or a dict of metric values. """
        raise NotImplementedError


class MetricWrapper(Metric):
    """ A wrapper class for easy-use of metric. """

    def __init__(self, flag, greater_is_better=True):
        super(MetricWrapper, self).__init__()
        self._flag = flag
        self._greater_is_better = greater_is_better

    def call(self, hypothesis, groundtruth=None):
        raise NotImplementedError("No need to call `__call__` in MetricWrapper.")

    def greater_or_eq(self, result1, result2):
        if self._greater_is_better:
            return super(MetricWrapper, self).greater_or_eq(result1, result2)
        return not super(MetricWrapper, self).greater_or_eq(result1, result2)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from neurst.data.text.moses_tokenizer import MosesTokenizer


def test():
    # lang, origin, tokenized, detokenized
    samples = [
        ("zh", False, [], "`啊你     好～！   ", "`啊你 好 ～ ！", "`啊你好 ～ ！"),
        ("en", False, [], "Hello p.m. 10, [.", "Hello p.m. 10 , [ .", None),
        ("en", True, [], "Hello p.m. 10, [.", "Hello p.m. 10 , [ .", None),
        ("en", False, ['<wotama>'], 'Hello <wotama> p.m. 10, [.',
         'Hello <wotama> p.m. 10 , [ .', 'Hello <wotama> p.m. 10, [.')

    ]

    for lang, escape, gloss, ori, tok, detok in samples:
        if not detok:
            detok = ori
        tokenizer = MosesTokenizer(language=lang, glossaries=gloss)
        assert tok == tokenizer.tokenize(ori, return_str=True)
        assert detok == tokenizer.detokenize(tok, return_str=True)


if __name__ == '__main__':
    test()



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import tensorflow as tf
from absl import logging

from neurst.optimizers.schedules import register_lr_schedule
from neurst.utils import compat
from neurst.utils.flags_core import Flag


@register_lr_schedule("noam")
class NoamSchedule(tf.keras.optimizers.schedules.LearningRateSchedule):
    """Learning rate schedule."""

    def __init__(self, args):
        """Initialize configuration of the learning rate schedule.

        Args:
          args: A dict of full parameters.
        """
        super(NoamSchedule, self).__init__()

        self._dmodel = args["dmodel"]
        self._warmup_steps = tf.cast(args["warmup_steps"], tf.float32)
        self._initial_step = compat.get_registered_initial_step()
        logging.info("Initialize NoamSchedule from global step={}. "
                     "The result learning rate will be scaled by {}"
                     "".format(int(self._initial_step), args["initial_factor"]))
        self._initial_step = tf.convert_to_tensor(self._initial_step, dtype=tf.float32)
        _initial_learning_rate = args["initial_factor"]
        self._initial_learning_rate = tf.convert_to_tensor(_initial_learning_rate, tf.float32)
        _end_learning_rate = args["end_factor"]
        if (_end_learning_rate is not None and args["start_decay_at"] is not None
            and args["decay_steps"] is not None):
            start_decay_at = args["start_decay_at"]
            decay_steps = args["decay_steps"]
            logging.info("\tThe scaling factor will start to decay from {} to {} at step {} "
                         "and finish at step {}.".format(_initial_learning_rate, _end_learning_rate,
                                                         start_decay_at, start_decay_at + decay_steps))
        else:
            _end_learning_rate = _initial_learning_rate
            start_decay_at = 0
            decay_steps = 1
        self._end_learning_rate = tf.convert_to_tensor(_end_learning_rate, tf.float32)
        self._start_decay_at = tf.convert_to_tensor(start_decay_at, tf.float32)
        self._decay_steps = tf.convert_to_tensor(decay_steps, tf.float32)

    @staticmethod
    def class_or_method_args():
        return [
            Flag("dmodel", dtype=Flag.TYPE.INTEGER, default=None,
                 help="The model dimension in the hidden layers."),
            Flag("warmup_steps", dtype=Flag.TYPE.INTEGER, default=4000,
                 help="The number of steps required for linear warmup."),
            Flag("initial_factor", dtype=Flag.TYPE.FLOAT, default=1.,
                 help="The initial learning rate scaling factor."),
            Flag("end_factor", dtype=Flag.TYPE.FLOAT, default=None,
                 help="The final decayed learning rate scaling factor."),
            Flag("start_decay_at", dtype=Flag.TYPE.INTEGER, default=0,
                 help="The `initial_factor` will start to decay at this step."),
            Flag("decay_steps", dtype=Flag.TYPE.INTEGER, default=None,
                 help="The `initial_factor` will decay to `end_learning_rate` in this many steps.")
        ]

    def __call__(self, global_step):
        """Calculate learning rate with linear warmup and rsqrt decay.

        Args:
          global_step: An integer, the current global step used for learning rate
            calculation.

        Returns:
          A float, the learning rate needs to be used for current global step.
        """
        with tf.name_scope('learning_rate_schedule'):
            global_step = tf.cast(global_step, tf.float32) + self._initial_step + 1.
            step_factor = tf.maximum(tf.minimum(
                global_step - self._start_decay_at, self._decay_steps), 0.)
            learning_rate = self._end_learning_rate + (
                self._initial_learning_rate - self._end_learning_rate) * (1. - step_factor / self._decay_steps)
            learning_rate *= (self._dmodel ** -0.5)
            # Apply linear warmup
            learning_rate *= tf.minimum(1.0, global_step / self._warmup_steps)
            # Apply rsqrt decay
            learning_rate /= tf.sqrt(tf.maximum(global_step, self._warmup_steps))
            return learning_rate

    def get_config(self):
        """Get the configuration of the learning rate schedule."""
        return {
            "initial_factor": float(self._initial_learning_rate.numpy()),
            "dmodel": self._dmodel,
            "warmup_steps": int(self._warmup_steps.numpy()),
            "end_factor": float(self._end_learning_rate.numpy()),
            "start_decay_at": int(self._start_decay_at.numpy()),
            "decay_steps": int(self._decay_steps.numpy())
        }



import tempfile

import tensorflow as tf

from neurst.data.data_pipelines.text_data_pipeline import TextDataPipeline


def test():
    vocab_file = tempfile.NamedTemporaryFile(delete=False)
    with tf.io.gfile.GFile(vocab_file.name, "w") as fw:
        for t in ["技术", "迅@@", "猛"]:
            fw.write(t + "\t100\n")
    bpe_codes_file = tempfile.NamedTemporaryFile(delete=False)
    with tf.io.gfile.GFile(bpe_codes_file.name, "w") as fw:
        fw.write("version\n")
        fw.write("\n".join(["技 术</w>", "发 展</w>"]) + "\n")
    text_dp = TextDataPipeline(
        vocab_path=vocab_file.name,
        language="zh",
        subtokenizer="bpe",
        subtokenizer_codes=bpe_codes_file.name)

    assert text_dp.process("技术 发展 迅猛") == [0, 3, 3, 1, 2, 5]
    assert text_dp.recover([0, 3, 1, 2, 5]) == "技术 <UNK> 迅猛"


if __name__ == "__main__":
    test()



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from abc import ABCMeta, abstractmethod

import six


@six.add_metaclass(ABCMeta)
class Criterion(object):
    REGISTRY_NAME = "criterion"

    def __init__(self):
        self._model = None

    @staticmethod
    def class_or_method_args():
        """ Returns a list of args for flag definition. """
        return []

    @abstractmethod
    def __call__(self, model_inp, model_out):
        """ Calculates according to model inputs and model outputs.

        Returns a list of tensors.
        """
        raise NotImplementedError

    @abstractmethod
    def reduce_loss(self, model_inp, model_out):
        """ Reduces loss tensor for training according to the model inputs
            and outputs.

        Returns: A float tensor.
        """
        raise NotImplementedError

    @abstractmethod
    def reduce_metrics(self, eval_res_list):
        """ Reduces the metrics according to a list of returned value from `eval`.

        Args:
            eval_res_list: A list of tuples of numpy.ndarray generated by `self.__call__`
                and model.__call__.

        Returns:
            A dict of reduced metrics for evaluation.
        """
        raise NotImplementedError

    @abstractmethod
    def as_metric(self):
        """ Returns a wrapper class of Metric. """
        raise NotImplementedError

    def set_model(self, model):
        self._model = model

    @property
    def model(self):
        return self._model



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import sys

import tensorflow as tf

from neurst.data.dataset_utils import glob_tfrecords


def cli_main():
    if len(sys.argv) == 1 or (len(sys.argv) == 2 and (sys.argv[1] in ["help", "--help", "-h"])):
        print("Usage: ")
        print("    >> python3 -m neurst.cli.view_tfrecord path")
        print("           Show examples and types of TF Record elements.")
        exit()

    print("===================== Examine elements =====================")
    for x in tf.data.TFRecordDataset(glob_tfrecords(sys.argv[1])).take(1):
        example = tf.train.Example()
        example.ParseFromString(x.numpy())
        print(example)
        print("elements: {")
        for name in example.features.feature:
            if len(example.features.feature[name].bytes_list.value) > 0:
                print(f"    \"{name}\": bytes (str)")
            elif len(example.features.feature[name].int64_list.value) > 0:
                print(f"    \"{name}\": int64")
            elif len(example.features.feature[name].float_list.value) > 0:
                print(f"    \"{name}\": float32")
        print("}")


if __name__ == "__main__":
    cli_main()



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re
import sys

import tensorflow as tf

from neurst.models.model_utils import _summary_model_variables
from neurst.utils.compat import wrapper_var_name


def cli_main():
    structured = False
    if "--structured" in sys.argv:
        structured = True
        sys.argv.remove("--structured")
    if len(sys.argv) == 1 or (len(sys.argv) == 2 and (sys.argv[1] in ["help", "--help", "-h"])):
        print("Usage: ")
        print("    >> python3 -m neurst.cli.inspect_checkpoint modeldir_or_checkpoint (--structured)")
        print("           List all variables and their shapes.")
        print()
        print("    >> python3 -m neurst.cli.inspect_checkpoint model_dir/checkpoint regular_expr")
        print("           List the variables and their shapes if the name matches the `regular_expr`.")
        print()
        print("    >> python3 -m neurst.cli.inspect_checkpoint model_dir/checkpoint var_name")
        print("           Print the variable tensor.")
        exit()

    model_dir = sys.argv[1]
    var_name = None
    if len(sys.argv) == 3:
        var_name = sys.argv[2]
    latest_ckpt_path = tf.train.latest_checkpoint(model_dir)
    if not latest_ckpt_path:
        latest_ckpt_path = model_dir
    try:
        varname_shape_list = tf.train.list_variables(latest_ckpt_path)
    except (tf.errors.NotFoundError, ValueError, tf.errors.DataLossError):
        print(f"ERROR: fail to load checkpoint from {model_dir}")
        exit()
    clean_varname2ckpt_varname = {
        wrapper_var_name(varname): (varname, shape)
        for varname, shape in varname_shape_list}
    specify_varname = False
    if var_name is not None:
        clean_varname2ckpt_varname = {
            wrapper_var_name(varname): (varname, shape)
            for varname, shape in varname_shape_list}
        if var_name in clean_varname2ckpt_varname:
            specify_varname = True
            print(f"Variable name: {var_name}")
            print(f"Tensor Shape: {str(clean_varname2ckpt_varname[var_name][1])}")
            print("Tensor Value: ")
            print(tf.train.load_variable(latest_ckpt_path, clean_varname2ckpt_varname[var_name][0]))
    if not specify_varname:
        if not structured:
            if var_name is None:
                print("\tvariable name \t shape")
            else:
                print(f"\tvariable name ({var_name}) \t shape")
        print_varname_shape_list = []
        for clean_varname, (varname, shape) in clean_varname2ckpt_varname.items():
            if varname in ["_CHECKPOINTABLE_OBJECT_GRAPH"]:
                continue
            if var_name is None or re.search(var_name, clean_varname):
                print_varname_shape_list.append((clean_varname, shape))
        if structured:
            _summary_model_variables(print_varname_shape_list, print)
        else:
            for clean_varname, shape in print_varname_shape_list:
                print(clean_varname + "\t" + str(shape))


if __name__ == "__main__":
    cli_main()



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from absl import logging

from neurst.data.text import register_tokenizer
from neurst.data.text.tokenizer import Tokenizer


@register_tokenizer
class Jieba(Tokenizer):

    def __init__(self, language="zh", glossaries=None):
        super(Jieba, self).__init__(
            language=language, glossaries=glossaries)
        if self._glossaries and len(self._glossaries) > 0:
            logging.info("WARNING: now `glossaries` has no effect on Jieba.")
        try:
            import jieba
            self._cut_fn = jieba.lcut
        except ImportError:
            raise ImportError('Please install jieba with: pip3 install jieba')

    def tokenize(self, text, return_str=False):
        return self._output_wrapper(self._cut_fn(self._convert_to_str(text)),
                                    return_str=return_str)

    def detokenize(self, text, return_str=True):
        return self._output_wrapper(
            self.cjk_deseg(self._convert_to_str(text)), return_str=return_str)



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from abc import ABCMeta, abstractmethod

import six
import tensorflow as tf
from absl import logging

from neurst.data.dataset_utils import load_tfrecords
from neurst.utils.compat import get_distributed_worker_setting
from neurst.utils.flags_core import Flag
from neurst.utils.misc import to_numpy_or_python_type


@six.add_metaclass(ABCMeta)
class Dataset(object):
    """ Abstract dataset class for handling data io. """
    REGISTRY_NAME = "dataset"

    def __init__(self):
        """ Initializes the dataset."""
        self._count = None

    @property
    @abstractmethod
    def status(self):
        raise NotImplementedError

    def build(self, auto_shard=False, map_func=None, map_output_dtypes=None,
              shuffle=True) -> tf.data.Dataset:
        """ Reads data from files and build the tf dataset.

        Args:
            auto_shard: Whether to automatically shard the dataset.
            map_func: A function mapping a dataset element to another dataset element.
            map_output_dtypes: A list/tuple of dtypes after applying `map_func`.
            shuffle: Whether to shuffle the TF records files.

        Returns: A tf.data.Dataset.
        """
        dataset = tf.data.Dataset.from_generator(self.build_iterator(map_func),
                                                 output_types=map_output_dtypes)
        if auto_shard:
            worker_id, num_workers, strategy = get_distributed_worker_setting()
            if num_workers > 1 and strategy in ["horovod", "byteps"]:
                logging.info("Shard %d of the whole dataset(total %d workers).", worker_id, num_workers)
                dataset = dataset.shard(num_workers, worker_id)
        return dataset

    @abstractmethod
    def build_iterator(self, map_func=None, shard_id=0, total_shards=1):
        """ Returns the iterator of the dataset.

        Args:
            map_func: A function mapping a dataset element to another dataset element.
            shard_id: Generator yields on the `shard_id`-th shard of the whole dataset.
            total_shards: The number of total shards.
        """
        raise NotImplementedError

    @property
    def num_samples(self):
        if self._count is None:
            cnt = 0
            for _ in self.build_iterator(map_func=None)():
                cnt += 1
            self._count = cnt
        return self._count


@six.add_metaclass(ABCMeta)
class TFRecordDataset(Dataset):
    """ Abstract dataset class for handling data io. """

    def __init__(self, args):
        super(TFRecordDataset, self).__init__()
        self._data_path = args["data_path"]
        self._shuffle_dataset = args["shuffle_dataset"]

    @staticmethod
    def class_or_method_args():
        return [
            Flag("data_path", dtype=Flag.TYPE.STRING, help="The path to TF records."),
            Flag("shuffle_dataset", dtype=Flag.TYPE.BOOLEAN,
                 help="Whether to shuffle the TF records files. "
                      "Note that some parts may be lost under MultiWorkerMirroredStrategy if set True."),
        ]

    @property
    @abstractmethod
    def fields(self) -> dict:
        """ The fields of the TF Records, e.g. {"feature": tf.io.VarLenFeature(tf.int64)}. """
        raise NotImplementedError

    @property
    @abstractmethod
    def status(self):
        raise NotImplementedError

    def build(self, auto_shard=False, map_func=None, map_output_dtypes=None,
              shuffle=True) -> tf.data.Dataset:
        """ Reads data from files and build the tf dataset.

        Args:
            auto_shard: Whether to automatically shard the dataset.
            map_func: A function mapping a dataset element to another dataset element.
            map_output_dtypes: A list/tuple of dtypes after applying `map_func`.
            shuffle: Whether to shuffle the TF records files.

        Returns: A tf.data.Dataset.
        """
        _ = map_output_dtypes
        return load_tfrecords(self._data_path, shuffle=self._shuffle_dataset and shuffle,
                              deterministic=(not shuffle),
                              auto_shard=auto_shard, map_func=map_func,
                              name_to_features=self.fields)

    def build_iterator(self, map_func=None, shard_id=0, total_shards=1):
        """ Returns the iterator of the dataset. """

        def gen():
            ds = load_tfrecords(self._data_path, shuffle=False, auto_shard=False,
                                name_to_features=self.fields,
                                sharding_index=shard_id, num_shards=total_shards)
            for x in ds:
                data = to_numpy_or_python_type(x, bytes_as_str=True)
                if map_func is not None:
                    data = map_func(data)
                yield data

        return gen



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import tensorflow as tf
from absl import logging

from neurst.data import dataset_utils
from neurst.tasks import register_task
from neurst.tasks.seq2seq import Seq2Seq
from neurst.training.training_utils import (EFFICIENT_MULTIPLIER, GPU_EFFICIENT_LEVEL, maximum_lower_multiple,
                                            minimal_multiple)
from neurst.utils import compat
from neurst.utils.configurable import deep_merge_dict
from neurst.utils.flags_core import Flag


def _auto_scale_batch_size(batch_size, level):
    if level == GPU_EFFICIENT_LEVEL.LEVEL0:
        return batch_size
    elif level == GPU_EFFICIENT_LEVEL.LEVEL1:
        return int(batch_size * 1.03)
    elif level == GPU_EFFICIENT_LEVEL.LEVEL2:
        return int(batch_size * 1.1)
    elif level == GPU_EFFICIENT_LEVEL.LEVEL3:
        return int(batch_size * 1.33)
    elif level == GPU_EFFICIENT_LEVEL.LEVEL4:
        return int(batch_size * 1.87)
    elif level == GPU_EFFICIENT_LEVEL.LEVEL5:
        return batch_size * 2


@register_task
class Translation(Seq2Seq):
    """ Defines the translation task. """

    @staticmethod
    def class_or_method_args():
        this_args = super(Translation, Translation).class_or_method_args()
        this_args.extend([
            Flag("gpu_efficient_level", dtype=Flag.TYPE.INTEGER, default=GPU_EFFICIENT_LEVEL.LEVEL1,
                 choices=tuple(GPU_EFFICIENT_LEVEL),
                 help="The efficient level for training using XLA, from 0~5."),
            Flag("auto_scaling_batch_size", dtype=Flag.TYPE.BOOLEAN, default=None,
                 help="Whether to automatically scale up the batch size to match the real tokens "
                      "when `gpu_efficient_level` > 0")
        ])
        return this_args

    def create_and_batch_tfds(self, ds, mode, args=None, num_replicas_in_sync=1):
        """ With efficient level for training. """
        if mode in [compat.ModeKeys.INFER, compat.ModeKeys.EVAL]:
            return super(Translation, self).create_and_batch_tfds(
                ds, mode, args, num_replicas_in_sync)
        if args is None:
            args = self._args
        else:
            args = deep_merge_dict(self._args, args, local_overwrite=False)
        level = args.get("gpu_efficient_level", None)
        auto_scale_batch = args.get("auto_scaling_batch_size", None)
        logging.info(f"Creating training dataset with GPU efficient level={level}.")
        dataset = ds.build(map_func=self.get_data_preprocess_fn(mode, ds.status, args),
                           map_output_dtypes=self.inputs_signature(mode)[0],
                           auto_shard=True, shuffle=True)
        dataset = dataset_utils.clean_dataset_by_length(
            dataset, {"feature": args["max_src_len"], "label": args["max_trg_len"]})
        if args["cache_dataset"]:
            dataset = dataset.cache()
        if args["shuffle_buffer"]:
            dataset = dataset.shuffle(buffer_size=args["shuffle_buffer"])
        padding_values = {"feature": tf.constant(self._src_data_pipeline.meta["eos_id"], dtype=tf.int64),
                          "label": tf.constant(self._trg_data_pipeline.meta["eos_id"], dtype=tf.int64)}
        if args["max_src_len"] is None:
            raise RuntimeError("Must provide `max_src_len` for training.")
        if args["max_trg_len"] is None:
            raise RuntimeError("Must provide `max_trg_len` for training.")
        max_src_len = minimal_multiple(args["max_src_len"], EFFICIENT_MULTIPLIER[level])
        max_trg_len = minimal_multiple(args["max_trg_len"], EFFICIENT_MULTIPLIER[level])
        max_len = max(max_src_len, max_trg_len)
        batch_size = dataset_utils.adjust_batch_size(args["batch_size"], args["batch_size_per_gpu"],
                                                     num_replicas_in_sync=num_replicas_in_sync,
                                                     verbose=False)
        if auto_scale_batch:
            batch_size = _auto_scale_batch_size(batch_size, level)
            logging.info(f"Auto scaling batch size to {batch_size}.")
        if level == GPU_EFFICIENT_LEVEL.LEVEL5:  # static batch
            _batch_size = batch_size
            if args["batch_by_tokens"]:
                _batch_size = _batch_size // max_len
            logging.info("Batching dataset with fixed shape: "
                         f"batch={_batch_size} x (feature={max_src_len}, label={max_trg_len}).")
            return dataset.padded_batch(
                _batch_size // num_replicas_in_sync * num_replicas_in_sync,
                padded_shapes={"feature": [max_src_len], "label": [max_trg_len]},
                drop_remainder=True, padding_values=padding_values)
        else:
            src_bucket_boundaries = [EFFICIENT_MULTIPLIER[level] * i for i in
                                     range(1, max_src_len // EFFICIENT_MULTIPLIER[level] + 1)]
            if src_bucket_boundaries[-1] < max_src_len:
                src_bucket_boundaries.append(minimal_multiple(src_bucket_boundaries[-1] + 1,
                                                              EFFICIENT_MULTIPLIER[level]))
            trg_bucket_boundaries = [EFFICIENT_MULTIPLIER[level] * i for i in
                                     range(1, max_trg_len // EFFICIENT_MULTIPLIER[level] + 1)]
            if trg_bucket_boundaries[-1] < max_trg_len:
                trg_bucket_boundaries.append(minimal_multiple(trg_bucket_boundaries[-1] + 1,
                                                              EFFICIENT_MULTIPLIER[level]))
            src_bucket_boundaries, trg_bucket_boundaries = dataset_utils.associated_bucket_boundaries(
                src_bucket_boundaries, trg_bucket_boundaries)
            bucket_boundaries = {
                "feature": src_bucket_boundaries,
                "label": trg_bucket_boundaries
            }
            bucket_batch_sizes = dataset_utils.adjust_batch_size(
                batch_size,
                bucket_boundaries=bucket_boundaries if args["batch_by_tokens"] else None,
                boundaries_reduce_to_length_fn=lambda x: max(tf.nest.flatten(x)),
                num_replicas_in_sync=num_replicas_in_sync)
            if level != GPU_EFFICIENT_LEVEL.LEVEL0:
                if isinstance(bucket_batch_sizes, list):
                    bucket_batch_sizes = [
                        int(maximum_lower_multiple(x // num_replicas_in_sync,
                                                   EFFICIENT_MULTIPLIER[level]) * num_replicas_in_sync)
                        for x in bucket_batch_sizes]
                else:
                    bucket_batch_sizes = int(maximum_lower_multiple(
                        bucket_batch_sizes // num_replicas_in_sync,
                        EFFICIENT_MULTIPLIER[level]) * num_replicas_in_sync)
            return dataset_utils.batch_examples_by_token(
                dataset,
                bucket_boundaries=bucket_boundaries,
                bucket_batch_sizes=bucket_batch_sizes,
                padding_values=padding_values,
                example_length_func=lambda x: {k: tf.size(v) for k, v in x.items()}
            )



import numpy
import tensorflow as tf

from neurst.layers.common_layers import (MultiHeadDenseLayer, PositionEmbeddingWrapper, PrePostProcessingWrapper,
                                         TransformerFFN)
from neurst.layers.modalities.text_modalities import WordEmbeddingSharedWeights


def test_ffn():
    ffn_layer = TransformerFFN(4, 3, 0.1, name="ffn")
    output = ffn_layer(tf.convert_to_tensor([[[1, 2.]]], dtype=tf.float32))
    assert output.get_shape() == tf.TensorShape([1, 1, 3])
    for w in ffn_layer.trainable_weights:
        if w.name == "ffn/dense1/kernel:0":
            assert w.get_shape() == tf.TensorShape([2, 4])
        elif w.name == "ffn/dense1/bias:0":
            assert w.get_shape() == tf.TensorShape([4, ])
        elif w.name == "ffn/dense2/kernel:0":
            assert w.get_shape() == tf.TensorShape([4, 3])
        elif w.name == "ffn/dense2/bias:0":
            assert w.get_shape() == tf.TensorShape([3, ])
        else:
            raise ValueError


def test_prepost():
    layer = TransformerFFN(4, 3, 0.1, name="ffn")
    prepost_layer = PrePostProcessingWrapper(
        layer, dropout_rate=0.1, name="lpp")
    output = prepost_layer(
        tf.convert_to_tensor([[1, 2, 3.]], dtype=tf.float32),
        is_training=True)
    assert output.get_shape() == tf.TensorShape([1, 3])


def test_multihead_dense():
    num_heads = 3
    output_size = 6
    non_out_layer = MultiHeadDenseLayer(
        output_size, num_heads, use_bias=True, is_output_transform=False,
        name="nonoutput_transform")
    inputs = tf.convert_to_tensor(numpy.random.randn(2, 3, 6), dtype=tf.float32)
    layer_out = non_out_layer(inputs)
    kernel, bias = None, None
    for w in non_out_layer.trainable_weights:
        if "kernel" in w.name:
            kernel = w
        else:
            bias = w
    manual_out = tf.einsum("abc,cd->abd", inputs, kernel) + bias
    manual_out = tf.reshape(manual_out, tf.concat(
        [tf.shape(manual_out)[:-1], [num_heads, output_size // num_heads]], axis=0))
    assert numpy.sum((manual_out.numpy() - layer_out.numpy()) ** 2) < 1e-9

    num_inputs_per_head = 5
    out_layer = MultiHeadDenseLayer(
        output_size, num_heads, use_bias=True, is_output_transform=True,
        name="output_transform")
    inputs = tf.convert_to_tensor(numpy.random.randn(
        1, 2, num_heads, num_inputs_per_head), dtype=tf.float32)
    layer_out = out_layer(inputs)
    kernel, bias = None, None
    for w in out_layer.trainable_weights:
        if "kernel" in w.name:
            kernel = w
        else:
            bias = w
    manual_out = tf.matmul(
        tf.reshape(inputs, tf.concat([tf.shape(inputs)[:-2], [-1]], 0)),
        kernel) + bias
    assert numpy.sum((manual_out.numpy() - layer_out.numpy()) ** 2) < 1e-9

    output_size1 = 9
    non_out_multi_layer = MultiHeadDenseLayer(
        [output_size, output_size1], num_heads, use_bias=True,
        is_output_transform=False, name="nonoutput_transform")
    inputs = tf.convert_to_tensor(numpy.random.randn(2, 3, 6), dtype=tf.float32)
    layer_out0, layer_out1 = non_out_multi_layer(inputs)
    kernel, bias = None, None
    for w in non_out_multi_layer.trainable_weights:
        if "kernel" in w.name:
            kernel = w
        else:
            bias = w
    manual_out = tf.einsum("abc,cd->abd", inputs, kernel) + bias
    manual_out0, manual_out1 = tf.split(
        manual_out, [output_size, output_size1], axis=-1)
    manual_out0 = tf.reshape(manual_out0, tf.concat(
        [tf.shape(manual_out0)[:-1], [num_heads, output_size // num_heads]], axis=0))
    manual_out1 = tf.reshape(manual_out1, tf.concat(
        [tf.shape(manual_out1)[:-1], [num_heads, output_size1 // num_heads]], axis=0))
    assert numpy.sum((manual_out0.numpy() - layer_out0.numpy()) ** 2) < 1e-9
    assert numpy.sum((manual_out1.numpy() - layer_out1.numpy()) ** 2) < 1e-9


def test_position_embedding():
    embedding_layer = WordEmbeddingSharedWeights(
        embedding_dim=5, vocab_size=10,
        share_softmax_weights=False)
    embedding_layer = PositionEmbeddingWrapper(
        timing="sinusoids",
        embedding_layer=embedding_layer,)
    inputs1d = tf.convert_to_tensor([4, 7, 8], tf.int32)
    inputs2d = tf.convert_to_tensor([[3, 1, 1, 1], [8, 1, 6, 4], [6, 6, 0, 5]], tf.int32)
    _ = embedding_layer(inputs2d)
    assert len(embedding_layer.get_weights()) == 1
    assert "emb/weights" in embedding_layer.trainable_weights[0].name
    embedding_layer.set_weights([numpy.array(
        [[-0.22683287, 0.20732224, -0.10953838, 0.15318757, -0.07203472],
         [0.48726183, 0.53683335, 0.38046378, -0.42776877, 0.51263684],
         [-0.20618078, 0.43871957, 0.26764846, 0.57276505, -0.13321346],
         [0.34847826, 0.1998071, 0.48136407, -0.03138721, -0.5397158],
         [-0.31466845, 0.24504018, 0.38156456, -0.03245735, 0.28105468],
         [-0.4769836, -0.2763745, -0.35024986, 0.5304734, -0.2523746],
         [0.13987714, -0.36480358, 0.5633767, 0.04371119, -0.5429846],
         [0.07482189, 0.4224295, 0.5645891, -0.12718052, 0.3637674],
         [0.4379062, 0.11231863, -0.6134181, -0.53932106, -0.5402442],
         [-0.18054467, -0.21964127, -0.14727849, 0.61441237, -0.13402274]])])

    emb_for_2d = embedding_layer(inputs2d)
    emb_for_1d = embedding_layer(inputs1d, time=3)
    assert numpy.sum((emb_for_2d.numpy() - numpy.array(
        [[[0.77922106, 0.4467823, 2.0763628, 0.92981607,
           -1.2068413],
          [1.9310216, 1.2004958, 1.3910451, 0.04347992,
           1.1462909],
          [1.998848, 1.2005959, 0.43459606, 0.04347992,
           1.1462909],
          [1.2306706, 1.2006959, -0.13924962, 0.04347986,
           1.1462909]],

         [[0.9791881, 0.2511521, -0.37164462, -0.2059586,
           -1.2080228],
          [1.9310216, 1.2004958, 1.3910451, 0.04347992,
           1.1462909],
          [1.2220722, -0.81552565, 0.8436019, 1.0977412,
           -1.2141505],
          [-0.56250006, 0.5482265, -0.13678819, 0.9274231,
           0.62845737]],

         [[0.3127748, -0.8157256, 2.2597487, 1.0977412,
           -1.2141505],
          [1.1542457, -0.8156256, 1.800051, 1.0977412,
           -1.2141505],
          [0.4020837, 0.46378663, -0.6610821, 1.3425379,
           -0.16107452],
          [-0.92544776, -0.6176922, -1.773175, 2.1861746,
           -0.56432676]]])) ** 2) < 1e-9
    assert numpy.sum((emb_for_1d.numpy() - numpy.array(
        [[-0.56250006, 0.5482265, -0.13678819, 0.9274231, 0.62845737],
         [0.30842686, 0.9448811, 0.27246714, 0.71561563, 0.8134086],
         [1.120308, 0.2514521, -2.361637, -0.20595866, -1.2080228]])) ** 2) < 1e-9

    emb_shared_layer = WordEmbeddingSharedWeights(
        embedding_dim=5, vocab_size=10,
        share_softmax_weights=True)
    emb_shared_layer = PositionEmbeddingWrapper(
        timing="emb",
        embedding_layer=emb_shared_layer)
    emb_for_2d = emb_shared_layer(inputs2d)
    logits_for_2d = emb_shared_layer(emb_for_2d, mode="linear")
    assert len(emb_shared_layer.get_weights()) == 3
    for w in emb_shared_layer.trainable_weights:
        if "shared/weights" in w.name:
            weights = w
        elif "shared/bias" in w.name:
            bias = w
    assert numpy.sum((numpy.dot(emb_for_2d.numpy(), numpy.transpose(weights.numpy()))
                      + bias.numpy() - logits_for_2d.numpy()) ** 2) < 1e-9


if __name__ == "__main__":
    test_ffn()
    test_prepost()
    test_multihead_dense()
    test_position_embedding()



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import tensorflow as tf
from absl import logging

from neurst.layers.common_layers import PositionEmbeddingWrapper
from neurst.layers.decoders import Decoder, build_decoder
from neurst.layers.encoders import Encoder, build_encoder
from neurst.layers.modalities.text_modalities import WordEmbeddingSharedWeights
from neurst.models import register_model
from neurst.models.model import BaseModel
from neurst.models.model_utils import input_length_to_padding
from neurst.utils.flags_core import Flag, ModuleFlag


@register_model(["seq2seq", "sequence_to_sequence", "SequenceToSequence"])
class EncoderDecoderModel(BaseModel):
    """ Defines the basic encoder-decoder model.

    All other encoder-decoder structure should inherit this class.
    """

    def __init__(self,
                 args,
                 src_meta,
                 trg_meta,
                 src_modality,
                 trg_modality,
                 encoder,
                 decoder,
                 name=None):
        """ Initializes a sequence to sequence model.

        Args:
            args: A dict, containing the model configuration.
            src_meta: A dict containing source-side vocabulary meta data, e.g. eos_id, unk_id.
            trg_meta: A dict containing target-side vocabulary meta data, e.g. eos_id, unk_id.
            src_modality: The source side modality.
            trg_modality: The target side modality.
            encoder: The encoder.
            decoder: The decoder.
            name: The name of the model.options = tf.data.Options()
        """
        super(EncoderDecoderModel, self).__init__(
            args=args, name=name or "SequenceToSequence")
        self._src_meta = src_meta
        self._trg_meta = trg_meta
        self._src_modality = src_modality
        self._trg_modality = trg_modality
        self._encoder = encoder
        self._decoder = decoder
        self._output_linear_layer = None
        if not self._args["modality.share_embedding_and_softmax_weights"]:
            self._output_linear_layer = tf.keras.layers.Dense(
                trg_meta["vocab_size"], activation=None,
                use_bias=True, name="softmax_linear")

    @staticmethod
    def class_or_method_args():
        return [
            ModuleFlag(Encoder.REGISTRY_NAME, default=None, help="The encoder."),
            ModuleFlag(Decoder.REGISTRY_NAME, default=None, help="The decoder."),
            Flag("modality.share_source_target_embedding", dtype=Flag.TYPE.BOOLEAN, default=False,
                 help="Whether to share source and target embedding table."),
            Flag("modality.share_embedding_and_softmax_weights", dtype=Flag.TYPE.BOOLEAN, default=False,
                 help="Whether to share the target embedding table and softmax weights."),
            Flag("modality.dim", dtype=Flag.TYPE.INTEGER, default=None,
                 help="The default embedding dimension for both source and target side."),
            Flag("modality.source.dim", dtype=Flag.TYPE.INTEGER, default=None,
                 help="The source-side embedding dimension, or `modality.dim` if not provided."),
            Flag("modality.target.dim", dtype=Flag.TYPE.INTEGER, default=None,
                 help="The target-side embedding dimension, or `modality.dim` if not provided."),
            Flag("modality.timing", dtype=Flag.TYPE.STRING, default=None,
                 help="The arbitrary parameters for positional encoding of both source and target side."),
            Flag("modality.source.timing", dtype=Flag.TYPE.STRING, default=None,
                 help="The arbitrary parameters for source-side positional encoding, "
                      "or `modality.timing` if not provided."),
            Flag("modality.target.timing", dtype=Flag.TYPE.STRING, default=None,
                 help="The arbitrary parameters for target-side positional encoding, "
                      "or `modality.timing` if not provided.")
        ]

    @classmethod
    def new(cls, args, src_meta, trg_meta, name=None):
        """ Builds a sequence to sequence model.

        Args:
            args: A dict containing all model parameters.
            src_meta: A dict containing source-side vocabulary meta data, e.g. eos_id, vocab_size.
            trg_meta: A dict containing target-side vocabulary meta data, e.g. eos_id, vocab_size.
            name: The name of the model.

        Returns:
            An encoder decoder model.
        """
        # build source and target modality
        src_modality, trg_modality = cls.build_modalities(args, src_meta, trg_meta)
        # build encoder and decoder
        encoder = build_encoder(args)
        decoder = build_decoder(args)
        model = cls(args, src_meta, trg_meta, src_modality, trg_modality, encoder, decoder, name=name)
        _ = model({"src": tf.convert_to_tensor([[1, 2, 3]], tf.int64),
                   "src_padding": tf.convert_to_tensor([[0, 0., 0]], tf.float32),
                   "trg_input": tf.convert_to_tensor([[1, 2, 3]], tf.int64)})
        return model

    @classmethod
    def build_modality(cls, vocab_size, emb_dim, name, timing=None,
                       share_embedding_and_softmax_weights=False):
        """ Creates modality layer.

        Args:
            vocab_size: An integer, the vocabulary size.
            emb_dim: An integer, the dimension of the embedding.
            timing: A string or a dict of parameter for positional embedding parameters.
            name: A string, the layer name.
            share_embedding_and_softmax_weights: Whether to share the embedding table and softmax weight.

        Returns:
            A modality layer.
        """
        modality = WordEmbeddingSharedWeights(
            embedding_dim=emb_dim, vocab_size=vocab_size,
            share_softmax_weights=share_embedding_and_softmax_weights,
            name=name)
        # position embedding wrapper
        if timing:
            if isinstance(timing, str):
                timing = {"timing": timing}
            elif not isinstance(timing, dict):
                raise ValueError("Unknown type of timing params: {}".format(str(timing)))
            modality = PositionEmbeddingWrapper(
                embedding_layer=modality, name=name + "_posenc_wrapper", **timing)
        return modality

    @classmethod
    def build_modalities(cls, model_args, src_meta, trg_meta):
        """ Create source and target modality. """
        # if modality.source.dim is not defined, then use modality.dim as default
        src_dim = model_args["modality.source.dim"] or model_args["modality.dim"]
        # if modality.target.dim is not defined, then use modality.dim as default
        trg_dim = model_args["modality.target.dim"] or model_args["modality.dim"]
        # whether to share source and target embedding
        if model_args["modality.share_source_target_embedding"]:
            assert src_meta["vocab_size"] == trg_meta["vocab_size"], (
                "Source vocab_size should be equal to target vocab_size "
                "when modality.share_source_and_target=True")
            logging.info("Ignore `modality.source.*` when sharing source and target modality.")
            input_name = target_name = "shared_symbol_modality"
        else:
            input_name = "input_symbol_modality"
            target_name = "target_symbol_modality"
        # creates target embedding table
        target_modality = cls.build_modality(
            vocab_size=trg_meta["vocab_size"], emb_dim=trg_dim, name=target_name,
            timing=(model_args["modality.target.timing"] or model_args["modality.timing"]),
            share_embedding_and_softmax_weights=model_args["modality.share_embedding_and_softmax_weights"])

        # creates source embedding table
        if model_args["modality.share_source_target_embedding"]:
            input_modality = target_modality
        else:
            input_modality = cls.build_modality(
                vocab_size=src_meta["vocab_size"], emb_dim=src_dim, name=input_name,
                timing=(model_args["modality.source.timing"] or model_args["modality.timing"]))

        return input_modality, target_modality

    def output_logits_layer(self, features):
        """ Projects the decoder output to logits. """
        if self._output_linear_layer is None:
            return self._trg_modality(features, mode="linear")
        else:
            return self._output_linear_layer(features)

    def get_symbols_to_logits_fn(self, inputs, is_training, is_inference,
                                 decode_padded_length=None):
        """ Prepares for decoding.

        Args:
            inputs: A dict of model inputs.
                - src: int tensor with shape [batch_size, src_input_length].
                - src_padding: float tensor with shape [batch_size, src_input_length].
                - trg_input: int tensor with shape [batch_size, trg_length].
            is_training: A bool, whether in training mode or not.
            is_inference: A bool, whether in generation mode or not.
            decode_padded_length: The maximum decoding length when inference, for creating
                static-shape cache.

        Returns:  A tuple of (decoding_internal_states, decoder_input, symbol_to_logit_fn)
        """
        embedded_inputs = self._src_modality(inputs["src"])
        src_padding = inputs.get("src_padding", None)
        if src_padding is None:
            src_padding = input_length_to_padding(inputs["src_length"], tf.shape(embedded_inputs)[1])
        encoder_outputs = self._encoder(embedded_inputs, src_padding, is_training=is_training)
        decoder_internal_cache = self._decoder.create_decoding_internal_cache(
            encoder_outputs=encoder_outputs,
            encoder_inputs_padding=src_padding,
            is_inference=is_inference,
            decode_padded_length=decode_padded_length)

        def symbols_to_logits_fn(symbols, cache, time=None):
            """ Generate logits for next potential IDs

            Args:
                symbols: Current decoded sequence.
                cache: A dictionary of values storing the encoder output, encoder-decoder
                    attention bias, and previous decoder attention values.
                time: Loop index, or None for transformer training

            Returns: The logits Tensor.
            """
            inputs = self._trg_modality(symbols, time=time)
            if decode_padded_length is None:
                decoder_output = self._decoder(inputs, cache, is_training=is_training,
                                               decode_loop_step=None)
            else:
                decoder_output = self._decoder(inputs, cache, is_training=is_training,
                                               decode_loop_step=time)
            logits = self.output_logits_layer(decoder_output)
            return logits

        generation_initializer = {
            "decoder_input": inputs["trg_input"],
            "decoder_internal_cache": decoder_internal_cache,
            "encoder_inputs_maxlen": tf.shape(encoder_outputs)[1],
            "eos_id": self._trg_meta["eos_id"],
            "unk_id": self._trg_meta["unk_id"]
        }
        return symbols_to_logits_fn, generation_initializer

    def call(self, inputs, is_training=True):
        """ Forward pass of the sequence to sequence model.

        Args:
            inputs: A dict of model inputs.
                - src: int tensor with shape [batch_size, src_input_length].
                - src_padding: float tensor with shape [batch_size, src_input_length].
                - trg_input: int tensor with shape [batch_size, trg_length].
            is_training: A bool, whether in training mode or not.

        Returns:
            A logits Tensor.
        """
        symbols_to_logits_fn, generation_initializer = self.get_symbols_to_logits_fn(
            inputs, is_training=is_training, is_inference=False)
        return symbols_to_logits_fn(generation_initializer["decoder_input"],
                                    generation_initializer["decoder_internal_cache"])



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" Base Encoder class. """
from __future__ import absolute_import, division, print_function

from abc import ABCMeta, abstractmethod

import six
import tensorflow as tf

from neurst.utils.configurable import extract_constructor_params


@six.add_metaclass(ABCMeta)
class Encoder(tf.keras.layers.Layer):
    """ Base class for encoders. """
    REGISTRY_NAME = "encoder"

    def __init__(self, name=None, **kwargs):
        """ Initializes the parameters of the encoders. """
        self._params = extract_constructor_params(locals(), verbose=False)
        super(Encoder, self).__init__(name=name)

    def build(self, input_shape):
        super(Encoder, self).build(input_shape)

    def get_config(self):
        return self._params

    @abstractmethod
    def call(self, inputs, inputs_padding, is_training=True):
        """ Encodes the inputs.

        Args:
            inputs: The embedded input, a float tensor with shape
                [batch_size, max_length, embedding_dim].
            inputs_padding: A float tensor with shape [batch_size, max_length],
                indicating the padding positions, where 1.0 for padding and
                0.0 for non-padding.
            is_training: A bool, whether in training mode or not.

        Returns:
            The encoded output with shape [batch_size, max_length, hidden_size]
        """
        raise NotImplementedError



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import tempfile

import tensorflow as tf

from neurst.data.text.bpe import BPE


def test():
    codes = ["技 术</w>", "发 展</w>"]
    tmp_file = tempfile.NamedTemporaryFile(delete=False)
    with tf.io.gfile.GFile(tmp_file.name, "w") as fw:
        fw.write("version\n")
        fw.write("\n".join(codes) + "\n")
    bpe = BPE(lang="zh",
              glossaries=["迅速", "<-neplhd-hehe>"])
    bpe.init_subtokenizer(tmp_file.name)

    tokens = bpe.tokenize("技术 发展 迅猛", return_str=True)
    assert tokens == "技术 发@@ 展 迅@@ 猛"
    assert bpe.detokenize(tokens) == "技术 发展 迅猛"
    tokens = bpe.tokenize("技术发展迅猛", return_str=True)
    assert tokens == "技@@ 术@@ 发@@ 展@@ 迅@@ 猛"
    assert bpe.detokenize(tokens) == "技术发展迅猛"
    tokens = bpe.tokenize("技术迅速发展迅速 迅速 <-neplhd-hehe>", return_str=True)
    assert tokens == "技术@@ 迅速@@ 发@@ 展@@ 迅速 迅速 <-neplhd-hehe>"
    assert bpe.detokenize(tokens) == "技术迅速发展迅速 迅速 <-neplhd-hehe>"

    os.remove(tmp_file.name)


if __name__ == "__main__":
    test()



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import tensorflow as tf

from neurst.layers.decoders import build_decoder
from neurst.layers.encoders import build_encoder
from neurst.models import register_model
from neurst.models.encoder_decoder_model import EncoderDecoderModel
from neurst.utils.flags_core import Flag
from neurst.utils.hparams_sets import register_hparams_set


@register_model
class LightConvolutionModel(EncoderDecoderModel):
    """ Defines the Transformer model.

    All other encoder-decoder structure should inherit this class.
    """

    def __init__(self, args, *largs, **kwargs):
        """ Initializes a sequence to sequence model.

        Args:
            args: A dict, containing the model configuration.
            src_meta: A dict containing source-side vocabulary meta data, e.g. eos_id, unk_id.
            trg_meta: A dict containing target-side vocabulary meta data, e.g. eos_id, unk_id.
            src_modality: The source side modality.
            trg_modality: The target side modality.
            encoder: The encoder.
            decoder: The decoder.
            name: The name of the model.options = tf.data.Options()
        """
        super(LightConvolutionModel, self).__init__(args, *largs, **kwargs)
        self._args = args

    @staticmethod
    def class_or_method_args():
        this_args = [x for x in super(LightConvolutionModel, LightConvolutionModel).class_or_method_args()
                     if x.name not in ["encoder.class", "decoder.class", "encoder.params", "decoder.params"]]
        this_args += [
            Flag("encoder.num_layers", dtype=Flag.TYPE.INTEGER, default=None,
                 help="The number of stacking layers of the encoder."),
            Flag("encoder.conv_kernel_size_list", dtype=Flag.TYPE.STRING, default=None,
                 help="A list of kernel sizes for each encoder layers. "
                      "The length of the list must be equal to `encoder.num_layers`."),
            Flag("encoder.num_conv_heads", dtype=Flag.TYPE.INTEGER, default=None,
                 help="The number of heads of encoder convolution shared weights."),
            Flag("encoder.conv_hidden_size", dtype=Flag.TYPE.INTEGER, default=None,
                 help="The number of hidden units of the encoder convolution layer."),
            Flag("encoder.conv_type", dtype=Flag.TYPE.STRING, default="lightweight",
                 help="The type of encoder conv layer, one of lightweight or dynamic."),
            Flag("encoder.filter_size", dtype=Flag.TYPE.INTEGER, default=None,
                 help="The number of the filter size of encoder ffn."),
            Flag("encoder.ffn_activation", dtype=Flag.TYPE.STRING, default="relu",
                 help="The activation function of encoder ffn layer."),
            Flag("encoder.conv_weight_dropout_rate", dtype=Flag.TYPE.FLOAT, default=0.,
                 help="The dropout rate of encoder convolution weights."),
            Flag("encoder.glu_after_proj", dtype=Flag.TYPE.BOOLEAN, default=False,
                 help="Whether to apply glu activation after input projection in encoder convolution layer."),
            Flag("encoder.ffn_dropout_rate", dtype=Flag.TYPE.FLOAT, default=0.,
                 help="The dropout rate of encoder ffn layer."),
            Flag("encoder.layer_postprocess_dropout_rate", dtype=Flag.TYPE.FLOAT, default=0.,
                 help="The dropout rate for each layer's post process in encoder."),
            Flag("encoder.layer_postprocess_epsilon", dtype=Flag.TYPE.FLOAT, default=1e-6,
                 help="The epsilon for layer normalization in encoder."),
            Flag("decoder.num_layers", dtype=Flag.TYPE.INTEGER, default=None,
                 help="The number of stacking layers of the decoder."),
            Flag("decoder.conv_kernel_size_list", dtype=Flag.TYPE.STRING, default=None,
                 help="A list of kernel sizes for each decoder layers. "
                      "The length of the list must be equal to `decoder.num_layers`."),
            Flag("decoder.num_conv_heads", dtype=Flag.TYPE.INTEGER, default=None,
                 help="The number of heads of decoder convolution shared weights."),
            Flag("decoder.conv_hidden_size", dtype=Flag.TYPE.INTEGER, default=None,
                 help="The number of hidden units of the decoder convolution layer."),
            Flag("decoder.num_attention_heads", dtype=Flag.TYPE.INTEGER, default=None,
                 help="The number of heads of decoder's encoder-decoder attention."),
            Flag("decoder.conv_type", dtype=Flag.TYPE.STRING, default="lightweight",
                 help="The type of decoder conv layer, one of lightweight or dynamic."),
            Flag("decoder.filter_size", dtype=Flag.TYPE.INTEGER, default=None,
                 help="The number of the filter size of decoder ffn."),
            Flag("decoder.ffn_activation", dtype=Flag.TYPE.STRING, default="relu",
                 help="The activation function of decoder ffn layer."),
            Flag("decoder.attention_dropout_rate", dtype=Flag.TYPE.FLOAT, default=0.,
                 help="The dropout rate of decoder's encoder-decoder attention."),
            Flag("decoder.conv_weight_dropout_rate", dtype=Flag.TYPE.FLOAT, default=0.,
                 help="The dropout rate of decoder convolution weights."),
            Flag("decoder.glu_after_proj", dtype=Flag.TYPE.BOOLEAN, default=False,
                 help="Whether to apply glu activation after input projection in decoder convolution layer."),
            Flag("decoder.attention_type", dtype=Flag.TYPE.STRING, default="dot_product",
                 help="The type of the attention function of decoder's encoder-decoder attention."),
            Flag("decoder.ffn_dropout_rate", dtype=Flag.TYPE.FLOAT, default=0.,
                 help="The dropout rate of decoder ffn layer."),
            Flag("decoder.layer_postprocess_dropout_rate", dtype=Flag.TYPE.FLOAT, default=0.,
                 help="The dropout rate for each layer's post process in decoder."),
            Flag("decoder.layer_postprocess_epsilon", dtype=Flag.TYPE.FLOAT, default=1e-6,
                 help="The epsilon for layer normalization in decoder."),
        ]
        return this_args

    @classmethod
    def new(cls, args, src_meta, trg_meta, name=None):
        """ Builds a sequence to sequence model.

        Args:
            args: A dict containing all model parameters.
            src_meta: A dict containing source-side vocabulary meta data, e.g. eos_id, vocab_size.
            trg_meta: A dict containing target-side vocabulary meta data, e.g. eos_id, vocab_size.
            name: The name of the model.

        Returns:
            An encoder decoder model.
        """
        # build source and target modality
        src_modality, trg_modality = cls.build_modalities(args, src_meta, trg_meta)
        encoder_params, decoder_params = {}, {}
        for f in cls.class_or_method_args():
            if f.name in args:
                if f.name.startswith("encoder."):
                    encoder_params[f.name[8:]] = args[f.name]
                elif f.name.startswith("decoder."):
                    decoder_params[f.name[8:]] = args[f.name]
        # build encoder and decoder
        encoder = build_encoder({
            "encoder.class": "LightConvolutionEncoder",
            "encoder.params": encoder_params})
        decoder = build_decoder({
            "decoder.class": "LightConvolutionDecoder",
            "decoder.params": decoder_params})
        model = cls(args, src_meta, trg_meta, src_modality, trg_modality, encoder, decoder, name=name)
        fake_inputs = {"src": tf.convert_to_tensor([[1, 2, 3]], tf.int64),
                       "src_padding": tf.convert_to_tensor([[0, 0., 0]], tf.float32),
                       "trg_input": tf.convert_to_tensor([[1, 2, 3]], tf.int64), }
        _ = model(fake_inputs)
        return model


def _common_hparams(dmodel, num_heads, filter_size,
                    encoder_layers, encoder_kernels,
                    decoder_layers, decoder_kernels,
                    encoder_conv_type, decoder_conv_type,
                    attention_dropout, weight_dropout, dropout):
    return {
        "model.class": "LightConvolutionModel",
        "model.params": {
            "modality.share_source_target_embedding": False,
            "modality.share_embedding_and_softmax_weights": True,
            "modality.dim": dmodel,
            "modality.timing": "sinusoids",
            "encoder.num_layers": encoder_layers,
            "encoder.conv_kernel_size_list": encoder_kernels,
            "encoder.num_conv_heads": num_heads,
            "encoder.conv_hidden_size": dmodel,
            "encoder.conv_type": encoder_conv_type,
            "encoder.filter_size": filter_size,
            "encoder.glu_after_proj": True,
            "encoder.conv_weight_dropout_rate": weight_dropout,
            "encoder.ffn_activation": "relu",
            "encoder.ffn_dropout_rate": attention_dropout,
            "encoder.layer_postprocess_dropout_rate": dropout,
            "decoder.num_layers": decoder_layers,
            "decoder.conv_kernel_size_list": decoder_kernels,
            "decoder.num_conv_heads": num_heads,
            "decoder.conv_hidden_size": dmodel,
            "decoder.num_attention_heads": num_heads,
            "decoder.conv_type": decoder_conv_type,
            "decoder.filter_size": filter_size,
            "decoder.attention_dropout_rate": attention_dropout,
            "decoder.glu_after_proj": True,
            "decoder.conv_weight_dropout_rate": weight_dropout,
            "decoder.attention_type": "dot_product",
            "decoder.ffn_activation": "relu",
            "decoder.ffn_dropout_rate": attention_dropout,
            "decoder.layer_postprocess_dropout_rate": dropout
        },
        "optimizer.class": "Adam",
        "optimizer.params": {
            "epsilon": 1.e-9,
            "beta_1": 0.9,
            "beta_2": 0.98
        },
        "lr_schedule.class": "noam",
        "lr_schedule.params": {
            "initial_factor": 2.0,
            "dmodel": dmodel,
            "warmup_steps": 10000
        },
    }


@register_hparams_set("lightweight_conv_big")
def lightweight_conv_big():
    return _common_hparams(
        dmodel=1024, num_heads=16, filter_size=4096,
        encoder_layers=7, encoder_kernels=[3, 7, 15, 31, 31, 31, 31],
        decoder_layers=6, decoder_kernels=[3, 7, 15, 31, 31, 31],
        encoder_conv_type="lightweight", decoder_conv_type="lightweight",
        attention_dropout=0.1, weight_dropout=0.1, dropout=0.3)


@register_hparams_set("lightweight_conv_big_dp01")
def lightweight_conv_big_dp01():
    return _common_hparams(
        dmodel=1024, num_heads=16, filter_size=4096,
        encoder_layers=7, encoder_kernels=[3, 7, 15, 31, 31, 31, 31],
        decoder_layers=6, decoder_kernels=[3, 7, 15, 31, 31, 31],
        encoder_conv_type="lightweight", decoder_conv_type="lightweight",
        attention_dropout=0.1, weight_dropout=0.1, dropout=0.1)


@register_hparams_set("lightweight_conv_toy")
def lightweight_conv_toy():
    return _common_hparams(
        dmodel=8, num_heads=4, filter_size=32,
        encoder_layers=2, encoder_kernels=[3, 7],
        decoder_layers=2, decoder_kernels=[3, 5],
        encoder_conv_type="lightweight", decoder_conv_type="lightweight",
        attention_dropout=0.1, weight_dropout=0.1, dropout=0.1)


@register_hparams_set("dynamic_conv_big")
def dynamic_conv_big():
    return _common_hparams(
        dmodel=1024, num_heads=16, filter_size=4096,
        encoder_layers=7, encoder_kernels=[3, 7, 15, 31, 31, 31, 31],
        decoder_layers=6, decoder_kernels=[3, 7, 15, 31, 31, 31],
        encoder_conv_type="dynamic", decoder_conv_type="dynamic",
        attention_dropout=0.1, weight_dropout=0.1, dropout=0.3)


@register_hparams_set("dynamic_conv_big_dp01")
def dynamic_conv_big_dp01():
    return _common_hparams(
        dmodel=1024, num_heads=16, filter_size=4096,
        encoder_layers=7, encoder_kernels=[3, 7, 15, 31, 31, 31, 31],
        decoder_layers=6, decoder_kernels=[3, 7, 15, 31, 31, 31],
        encoder_conv_type="dynamic", decoder_conv_type="dynamic",
        attention_dropout=0.1, weight_dropout=0.1, dropout=0.1)


@register_hparams_set("dynamic_conv_toy")
def dynamic_conv_toy():
    return _common_hparams(
        dmodel=8, num_heads=4, filter_size=32,
        encoder_layers=2, encoder_kernels=[3, 7],
        decoder_layers=2, decoder_kernels=[3, 5],
        encoder_conv_type="dynamic", decoder_conv_type="dynamic",
        attention_dropout=0.1, weight_dropout=0.1, dropout=0.1)



import tensorflow as tf
from absl import logging

from neurst.data.text import register_tokenizer
from neurst.data.text.tokenizer import Tokenizer


@register_tokenizer
class BPE(Tokenizer):

    def __init__(self,
                 glossaries=None,
                 separator='@@',
                 vocabulary=None,
                 version=(0, 2),
                 **kwargs):
        _ = kwargs
        super(BPE, self).__init__(language=None, glossaries=glossaries)
        self.glossaries = glossaries if glossaries else []
        self.version = version
        # some hacking to deal with duplicates (only consider first instance)

        self.separator = separator
        if vocabulary:
            with tf.io.gfile.GFile(vocabulary) as fp:
                self.vocab = [line.strip().split()[0] for line in fp]
        else:
            self.vocab = []
        self._built = False

    def init_subtokenizer(self, codes):
        if isinstance(codes, str):
            with tf.io.gfile.GFile(codes, "r") as fp:
                codes = [line.strip() for line in fp][1:]
        elif not isinstance(codes, list):
            raise ValueError("Not supported type of codes: {}.".format(type(codes)))
        if codes[0].startswith("#version"):
            codes = codes[1:]
        self.bpe_codes = [tuple(item.split()) for item in codes]
        self.bpe_codes = dict([(code, i) for (i, code) in reversed(list(enumerate(self.bpe_codes)))])
        self.bpe_codes_reverse = dict([(pair[0] + pair[1], pair) for pair, i in self.bpe_codes.items()])
        self._built = True

    @staticmethod
    def isolate_glossary(word, glossary):
        """
        Isolate a glossary present inside a word.

        Returns a list of subwords. In which all 'glossary' glossaries are isolated

        For example, if 'USA' is the glossary and '1934USABUSA' the word, the return value is:
            ['1934', 'USA', 'B', 'USA']
        """
        if word == glossary or glossary not in word:
            return [word]
        else:
            splits = word.split(glossary)
            segments = [segment.strip() for split in splits[:-1] for segment in [split, glossary] if segment != '']
            return segments + [splits[-1].strip()] if splits[-1] != '' else segments

    def _isolate_glossaries(self, word):
        word_segments = [word]
        for gloss in self.glossaries:
            word_segments = [out_segments for segment in word_segments
                             for out_segments in self.isolate_glossary(segment, gloss)]
        return word_segments

    @staticmethod
    def get_pairs(word):
        """Return set of symbol pairs in a word.

        word is represented as tuple of symbols (symbols being variable-length strings)
        """
        pairs = set()
        prev_char = word[0]
        for char in word[1:]:
            pairs.add((prev_char, char))
            prev_char = char
        return pairs

    @staticmethod
    def recursive_split(segment, bpe_codes, vocab, separator, final=False):
        """Recursively split segment into smaller units (by reversing BPE merges)
        until all units are either in-vocabulary, or cannot be split futher."""

        try:
            if final:
                left, right = bpe_codes[segment + '</w>']
                right = right[:-4]
            else:
                left, right = bpe_codes[segment]
        except KeyError:
            # sys.stderr.write('cannot split {0} further.\n'.format(segment))
            yield segment
            return

        if left + separator in vocab:
            yield left
        else:
            for item in BPE.recursive_split(left, bpe_codes, vocab, separator, False):
                yield item

        if (final and right in vocab) or (not final and right + separator in vocab):
            yield right
        else:
            for item in BPE.recursive_split(right, bpe_codes, vocab, separator, final):
                yield item

    @staticmethod
    def check_vocab_and_split(orig, bpe_codes, vocab, separator):
        """Check for each segment in word if it is in-vocabulary,
        and segment OOV segments into smaller units by reversing the BPE merge operations"""

        out = []

        for segment in orig[:-1]:
            if segment + separator in vocab:
                out.append(segment)
            else:
                # sys.stderr.write('OOV: {0}\n'.format(segment))
                for item in BPE.recursive_split(segment, bpe_codes, vocab, separator, False):
                    out.append(item)

        segment = orig[-1]
        if segment in vocab:
            out.append(segment)
        else:
            # sys.stderr.write('OOV: {0}\n'.format(segment))
            for item in BPE.recursive_split(segment, bpe_codes, vocab, separator, True):
                out.append(item)

        return out

    @staticmethod
    def bpe_encode(orig, bpe_codes, bpe_codes_reverse, vocab, separator, version, glossaries=None, cache={}):
        """Encode word based on list of BPE merge operations, which are applied consecutively
        """

        if orig in cache:
            return cache[orig]

        if orig in glossaries:
            cache[orig] = (orig,)
            return (orig,)

        if version == (0, 1):
            word = tuple(orig) + ('</w>',)
        elif version == (0, 2):  # more consistent handling of word-final segments
            word = tuple(orig[:-1]) + (orig[-1] + '</w>',)
        else:
            raise NotImplementedError

        pairs = BPE.get_pairs(word)

        if not pairs:
            return orig

        while True:
            bigram = min(pairs, key=lambda pair: bpe_codes.get(pair, float('inf')))
            if bigram not in bpe_codes:
                break
            first, second = bigram
            new_word = []
            i = 0
            while i < len(word):
                try:
                    j = word.index(first, i)
                    new_word.extend(word[i:j])
                    i = j
                except ValueError:
                    new_word.extend(word[i:])
                    break

                if word[i] == first and i < len(word) - 1 and word[i + 1] == second:
                    new_word.append(first + second)
                    i += 2
                else:
                    new_word.append(word[i])
                    i += 1
            new_word = tuple(new_word)
            word = new_word
            if len(word) == 1:
                break
            else:
                pairs = BPE.get_pairs(word)

        # don't print end-of-word symbols
        if word[-1] == '</w>':
            word = word[:-1]
        elif word[-1].endswith('</w>'):
            word = word[:-1] + (word[-1].replace('</w>', ''),)

        if vocab:
            word = BPE.check_vocab_and_split(word, bpe_codes_reverse, vocab, separator)

        cache[orig] = word
        return word

    def tokenize(self, text, return_str=False):
        if not self._built:
            raise ValueError("call `init_subtokenizer` at first to initialize the BPE.")
        try:
            tokens = self._convert_to_list(text)
            output = []
            for word in tokens:
                new_word = [out for segment in self._isolate_glossaries(word)
                            for out in BPE.bpe_encode(segment,
                                                      self.bpe_codes,
                                                      self.bpe_codes_reverse,
                                                      self.vocab,
                                                      self.separator,
                                                      self.version,
                                                      self.glossaries)]
                for item in new_word[:-1]:
                    output.append(item + self.separator)
                output.append(new_word[-1])
        except IndexError:
            logging.info(tokens)
            raise IndexError("string index out of range, string", tokens)
        return self._output_wrapper(output, return_str)

    @staticmethod
    def recov_bpe(text, seperator="@@"):
        """ Recovers BPE result.

        Args:
            text: A string.
            seperator:

        Returns:
            The recovered string.
        """
        text = Tokenizer._convert_to_str(text, delimiter=" ").replace(seperator + " ", "").strip()
        if text.endswith("@@"):
            text = text[:-2]
        return text

    def detokenize(self, text, return_str=True):
        return self._output_wrapper(
            self.recov_bpe(text, seperator=self.separator), return_str)



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" Implements light convolution encoder as described in https://arxiv.org/pdf/1901.10430.pdf. """
import tensorflow as tf

from neurst.layers import build_transformer_component
from neurst.layers.attentions.light_convolution_layer import LightConvolutionLayer
from neurst.layers.common_layers import TransformerFFN
from neurst.layers.encoders import register_encoder
from neurst.layers.encoders.encoder import Encoder


@register_encoder
class LightConvolutionEncoder(Encoder):
    """ Defines light convolution encoder as described https://arxiv.org/pdf/1901.10430.pdf. """

    def __init__(self,
                 num_layers,
                 conv_kernel_size_list,
                 num_conv_heads,
                 conv_hidden_size,
                 filter_size,
                 conv_type="lightweight",
                 glu_after_proj=True,
                 conv_weight_dropout_rate=0.,
                 ffn_activation="relu",
                 ffn_dropout_rate=0.,
                 layer_postprocess_dropout_rate=0.,
                 layer_postprocess_epsilon=1e-6,
                 name=None):
        """ Initializes the transformer encoders.

        Args:
            num_layers: The number of stacked layers.
            conv_kernel_size_list: An int list of encoder kernel sizes. The length of the list must
                be equal to `num_layers`.
            num_conv_heads: An integer, the number of heads for conv shared weights.
            conv_hidden_size: The hidden size of conv layer.
            filter_size: The filter size of ffn layer.
            conv_type: The type of conv layer, one of lightweight or dynamic.
            ffn_activation: The activation function of ffn layer.
            ffn_dropout_rate: The dropout rate for ffn layer.
            glu_after_proj: Whether to apply glu activation after input projection.
            conv_weight_dropout_rate: The dropout rate of the conv weights.
            layer_postprocess_dropout_rate: The dropout rate for each layer post process.
            layer_postprocess_epsilon: The epsilon for layer norm.
            name: The name of this encoder.
        """
        super(LightConvolutionEncoder, self).__init__(
            num_layers=num_layers, conv_kernel_size_list=conv_kernel_size_list,
            num_conv_heads=num_conv_heads, conv_hidden_size=conv_hidden_size,
            filter_size=filter_size, ffn_activation=ffn_activation,
            ffn_dropout_rate=ffn_dropout_rate, conv_type=conv_type,
            glu_after_proj=glu_after_proj,
            conv_weight_dropout_rate=conv_weight_dropout_rate,
            layer_postprocess_dropout_rate=layer_postprocess_dropout_rate,
            layer_postprocess_epsilon=layer_postprocess_epsilon,
            name=name or self.__class__.__name__)
        self._stacking_layers = []

    def build(self, input_shape):
        """ Builds the transformer encoder layer. """
        params = self.get_config()
        for lid in range(params["num_layers"]):
            self._stacking_layers.append([
                build_transformer_component({
                    "base_layer.class": LightConvolutionLayer.__name__,
                    "base_layer.params": dict(
                        kernel_size=params["conv_kernel_size_list"][lid],
                        num_heads=params["num_conv_heads"],
                        conv_type=params["conv_type"],
                        conv_dim=params["conv_hidden_size"],
                        use_glu=params["glu_after_proj"],
                        weight_dropout_rate=params["conv_weight_dropout_rate"],
                        name="light_conv"
                    )},
                    dropout_rate=params["layer_postprocess_dropout_rate"],
                    epsilon=params["layer_postprocess_epsilon"]),
                build_transformer_component({
                    "base_layer.class": TransformerFFN.__name__,
                    "base_layer.params": dict(
                        filter_size=params["filter_size"],
                        output_size=input_shape[-1],
                        dropout_rate=params["ffn_dropout_rate"],
                        activation=params["ffn_activation"],
                        name="ffn")},
                    dropout_rate=params["layer_postprocess_dropout_rate"],
                    epsilon=params["layer_postprocess_epsilon"])
            ])
        self._output_norm_layer = tf.keras.layers.LayerNormalization(
            epsilon=params["layer_postprocess_epsilon"],
            dtype="float32", name="output_ln")
        super(LightConvolutionEncoder, self).build(input_shape)

    def call(self, inputs, inputs_padding, is_training=True):
        """ Encodes the inputs.

        Args:
            inputs: The embedded input, a float tensor with shape
                [batch_size, max_length, embedding_dim].
            inputs_padding: A float tensor with shape [batch_size, max_length],
                indicating the padding positions, where 1.0 for padding andtf
                0.0 for non-padding.
            is_training: A bool, whether in training mode or not.

        Returns:
            The encoded output with shape [batch_size, max_length, hidden_size]
        """
        # [batch_size, max_length], FLOAT_MIN for padding, 0.0 for non-padding
        x = inputs
        if is_training:
            x = tf.nn.dropout(x, rate=self.get_config()[
                "layer_postprocess_dropout_rate"])
        for idx, layer in enumerate(self._stacking_layers):
            conv_layer = layer[0]
            ffn_layer = layer[1]
            with tf.name_scope("layer_{}".format(idx)):
                # self attention layer
                x = conv_layer(x, inputs_padding=inputs_padding, is_training=is_training)
                # ffn
                x = ffn_layer(x, is_training=is_training)

        return self._output_norm_layer(x)



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import math

import numpy
import tensorflow as tf
import yaml
from absl import logging


class SpecAugment(object):
    """ Implements SpecAugment, https://arxiv.org/abs/1904.08779 """

    _PREDEF_SETTINGS = {
        # LibriSpeech basic
        "LB": {
            "time_wrap_w": 80,
            "freq_mask_n": 1,
            "freq_mask_f": 27,
            "time_mask_n": 1,
            "time_mask_t": 100,
            "time_mask_p": 1.
        },
        # LibriSpeech double
        "LD": {
            "time_wrap_w": 80,
            "freq_mask_n": 2,
            "freq_mask_f": 27,
            "time_mask_n": 2,
            "time_mask_t": 100,
            "time_mask_p": 1.
        },
        # Switchboard mild
        "SM": {
            "time_wrap_w": 40,
            "freq_mask_n": 2,
            "freq_mask_f": 15,
            "time_mask_n": 2,
            "time_mask_t": 70,
            "time_mask_p": 0.2
        },
        # Switchboard strong
        "SS": {
            "time_wrap_w": 40,
            "freq_mask_n": 2,
            "freq_mask_f": 27,
            "time_mask_n": 2,
            "time_mask_t": 70,
            "time_mask_p": 0.2
        }
    }

    def __init__(self,
                 time_wrap_w: int,
                 freq_mask_n: int,
                 freq_mask_f: int,
                 time_mask_n: int,
                 time_mask_t: int,
                 time_mask_p: float,
                 mask_value=None):
        """ Initializes spec augment.

        Args:
            time_wrap_w: The time wrapping parameter.
            freq_mask_n: The number of times to apply frequency masking.
            freq_mask_f: The maximum consecutive frequency to be masked.
            time_mask_n: The number of times to apply time masking.
            time_mask_t: The maximum consecutive time steps to be masked.
            time_mask_p: The upper bound ratio of consecutive time steps of time masking.
            mask_value: The mask value, use the mean of spectrogram if not provided.
        """
        self._time_wrap_w = time_wrap_w
        self._freq_mask_n = freq_mask_n
        self._freq_mask_f = freq_mask_f
        self._time_mask_n = time_mask_n
        self._time_mask_t = time_mask_t
        self._time_mask_p = time_mask_p
        self._mask_value = mask_value
        assert self._time_mask_t > 0
        assert self._freq_mask_f > 0
        if self._time_wrap_w > 0:
            logging.info("WARNING(specaug): here the time wrapping is not implemented "
                         "because it is not a major factor in improving the performance "
                         "according to the original paper, and slows down the training speed.")

    @classmethod
    def build(cls, setting):
        if setting is None:
            return None
        setting = yaml.load(setting, Loader=yaml.FullLoader)
        if isinstance(setting, str):
            setting = cls._PREDEF_SETTINGS.get(setting, None)
        if setting is None:
            return None
        assert isinstance(setting, dict), f"Unknown type of setting: {setting}"
        return cls(**setting)

    @staticmethod
    def freq_or_time_masking_numpy(spectrogram, n, F,
                                   mask_value, axis, p=None):
        """ Mask the spectrogram.

        Repeat n times:
            f ~ [0, F)
            f0 ~ [0, nu - f)
            The [f0, f0 + f) positions of `spectrogram` on `axis` are overwritten by `mask_value`.

        Args:
            spectrogram: numpy.ndarray, the audio feature map, of shape [nframes, ndim].
            n: The number of repeated times.
            F: The maximum mask range.
            mask_value: The value to overwrite the masked positions.
            axis: The masked axis of `spectrogram`.
            p: The upper bound ratio of the mask range if provided.

        Returns:
            The augmented spectrogram.
        """
        num_frames_or_freqs = spectrogram.shape[axis]
        if num_frames_or_freqs < F:
            return spectrogram
        if p:
            F = min(F, math.floor(num_frames_or_freqs * p))
        # sample mask range
        f = numpy.random.randint(0, F, size=n)
        f0 = numpy.random.randint(0, num_frames_or_freqs - f)
        for i in range(n):
            if f0[i] == 0:
                continue
            if axis == 0:
                spectrogram[f0[i]: f0[i] + f[i], :] = mask_value
            else:
                assert axis == 1
                spectrogram[:, f0[i]: f0[i] + f[i]] = mask_value
        return spectrogram

    @staticmethod
    def freq_or_time_mask_tf(spectrogram, n, F, axis, p=None):
        """ Mask the spectrogram.

        Repeat n times:
            f ~ [0, F)
            f0 ~ [0, nu - f)
            The [f0, f0 + f) positions of `spectrogram` on `axis` are overwritten by `mask_value`.

        Args:
            spectrogram: numpy.ndarray, the audio feature map, of shape [nframes, ndim].
            n: The number of repeated times.
            F: The maximum mask range.
            axis: The masked axis of `spectrogram`.
            p: The upper bound ratio of the mask range if provided.

        Returns:
            The mask vector on `axis`.
        """
        num_frames_or_freqs = tf.shape(spectrogram)[axis]

        def _apply(n, F, p):
            # F = tf.cond(tf.less(num_frames_or_freqs, F * 2 + 1), lambda: num_frames_or_freqs // n, lambda: F)
            if p:
                F = tf.minimum(F, tf.math.floor(num_frames_or_freqs * p))
            masks = []
            # sample mask range
            for _ in range(n):
                f = tf.random.uniform(shape=[], minval=0, maxval=F, dtype=tf.int32)
                f0 = tf.random.uniform(shape=[], minval=0, maxval=num_frames_or_freqs - f, dtype=tf.int32)
                mask = tf.pad(tf.ones([f, ], dtype=tf.bool), paddings=[[f0, num_frames_or_freqs - f - f0]],
                              mode='CONSTANT', constant_values=False)
                masks.append(mask)
            final_mask = tf.reduce_any(tf.convert_to_tensor(masks, dtype=tf.bool), axis=0)
            return tf.cast(final_mask, spectrogram.dtype)

        return tf.cond(tf.less(num_frames_or_freqs, F),
                       lambda: tf.zeros([num_frames_or_freqs, ], dtype=spectrogram.dtype),
                       lambda: _apply(n, F, p))

    def _call_tf(self, spectrogram, true_length=None):
        """ Applies specaug based on TensorFlow operations.

        Args:
            spectrogram: A tf.Tensor of shape [nframes, nfeatures] or [batch, nframes, nfeatures].
            true_length: A tf.Tensor of shape [batch, ] if `spectrogram` is a tensor of shape
                [batch, nframes, nfeatures], else None.

        Returns:
            A tuple (augmented spectrogram, new true_length) if ndims of `spectrogram` is 3,
            else the augmented spectrogram.
        """
        _ = true_length
        ndim = spectrogram.get_shape().ndims
        if ndim == 2:
            time_mask = SpecAugment.freq_or_time_mask_tf(
                spectrogram, self._time_mask_n, self._time_mask_t, axis=0)
            freq_mask = SpecAugment.freq_or_time_mask_tf(
                spectrogram, self._freq_mask_n, self._freq_mask_f, axis=1)
            demask = tf.einsum("M,N->MN", 1 - time_mask, 1 - freq_mask)
            mask_value = self._mask_value
            if mask_value is None:
                mask_value = tf.reduce_mean(spectrogram)
            return spectrogram * demask + (1. - demask) * mask_value  # mask_value
        assert ndim == 3  # [batch, frames, dim]
        raise NotImplementedError

    def _call_numpy(self, spectrogram):
        """ Applies specaug based on numpy operations.

        Args:
            spectrogram: A numpy.ndarray of shape [nframes, nfeatures]

        Returns:
            The augmented spectrogram.
        """
        ndim = spectrogram.ndim
        if ndim > 2:
            raise ValueError("batch specaug is not implemented for numpy array.")
        mask_value = self._mask_value
        if mask_value is None:
            mask_value = spectrogram.mean()
        if self._freq_mask_n > 0:
            spectrogram = self.freq_or_time_masking_numpy(
                spectrogram, n=self._freq_mask_n, F=self._freq_mask_f,
                mask_value=mask_value, axis=1)
        if self._time_mask_n > 0:
            spectrogram = self.freq_or_time_masking_numpy(
                spectrogram, n=self._time_mask_n, F=self._time_mask_t,
                mask_value=mask_value, axis=0, p=self._time_mask_p)
        return spectrogram

    def __call__(self, spectrogram, true_length=None):
        """ Applies specaug.

        Args:
            spectrogram: A numpy.ndarray of shape [nframes, nfeatures]
                or a tf.Tensor of shape [nframes, nfeatures] or [batch, nframes, nfeatures].
            true_length: A tf.Tensor of shape [batch, ] if `spectrogram` is a tensor of shape
                [batch, nframes, nfeatures], else None.

        Returns:
            A tuple (augmented spectrogram, new true_length) if ndims of `spectrogram` is 3,
            else the augmented spectrogram.
        """
        if isinstance(spectrogram, tf.Tensor):
            return self._call_tf(spectrogram)
        return self._call_numpy(spectrogram)



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import numpy

from neurst.data.audio import FeatureExtractor, register_feature_extractor
from neurst.utils.flags_core import Flag

try:
    from python_speech_features import logfbank
except ImportError:
    pass


@register_feature_extractor("fbank")
class LogMelFbank(FeatureExtractor):

    def __init__(self, args):
        self._nfilt = args["nfilt"]
        self._winlen = args["winlen"]
        self._winstep = args["winstep"]
        try:
            from python_speech_features import logfbank
            _ = logfbank
        except ImportError:
            raise ImportError('Please install python_speech_features with: pip3 install python_speech_features')

    @staticmethod
    def class_or_method_args():
        return [
            Flag("nfilt", dtype=Flag.TYPE.INTEGER, default=80,
                 help="The number of frames in the filterbank."),
            Flag("winlen", dtype=Flag.TYPE.FLOAT, default=0.025,
                 help="The length of the analysis window in seconds. Default is 0.025s."),
            Flag("winstep", dtype=Flag.TYPE.FLOAT, default=0.01,
                 help="The step between successive windows in seconds. Default is 0.01s.")
        ]

    def __call__(self, signal, rate):
        inp = logfbank(signal, samplerate=rate, nfilt=self._nfilt,
                       winlen=self._winlen, winstep=self._winstep).astype(numpy.float32)
        inp = (inp - numpy.mean(inp)) / numpy.std(inp)
        return inp
